from datetime import datetime
from enum import Enum

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import ENUM
from sqlalchemy.ext.declarative import as_declarative

SCHEMA_NAME = "asset_info"


symbols_table = sa.Table(
    "symbols",
    sa.MetaData(schema=SCHEMA_NAME),
    sa.Column('data_id', sa.String, primary_key=True),
    sa.Column("symbol", sa.String),
    sa.Column("stock", sa.Boolean, comment="DataIDIgnore"),
    sa.Column("options", sa.Boolean, comment="DataIDIgnore"),
    sa.Column("futures", sa.Boolean, comment="DataIDIgnore"),
    sa.Column("top_liquid_stock", sa.Boolean, comment="DataIDIgnore"),
    sa.Column("top_liquid_option", sa.Boolean, comment="DataIDIgnore"),
    sa.Column("top_liquid_future", sa.Boolean, comment="DataIDIgnore"),
)


class CompanyIDType(Enum):
    SYMBOL = "SYMBOL"
    CIK = "CIK"
    NAME = "NAME"


@as_declarative()
class AssetInfo:
    __table_args__ = {"schema": SCHEMA_NAME}
    data_id = sa.Column(sa.String, primary_key=True)
    asset_id = sa.Column(sa.String)
    asset_id_type = sa.Column(
        ENUM(CompanyIDType, schema=SCHEMA_NAME), default=CompanyIDType.SYMBOL
    )


class NameType(Enum):
    COMPANY = "COMPANY"
    ASSET = "ASSET"
    UNKNOWN = "UNKNOWN"


class Name(AssetInfo):
    __tablename__ = "name"
    name = sa.Column(sa.String)
    name_type = sa.Column(ENUM(NameType, schema=SCHEMA_NAME), default=NameType.UNKNOWN)


class Type(AssetInfo):
    __tablename__ = "type"
    type = sa.Column(sa.String)


class CompanyFormerName(AssetInfo):
    __tablename__ = "company_former_name"
    company_former_name = sa.Column(sa.String)


class CIK(AssetInfo):
    __tablename__ = "cik"
    cik = sa.Column(sa.String, sa.CheckConstraint("cik ~ '\d{1,10}'", name="valid_cik"))


class SIC(AssetInfo):
    __tablename__ = "sic"
    sic = sa.Column(sa.String)


class Tag(AssetInfo):
    __tablename__ = "tag"
    tag = sa.Column(sa.String)


class Exchange(AssetInfo):
    __tablename__ = "exchange"
    exchange = sa.Column(sa.String)


class IPODate(AssetInfo):
    __tablename__ = "ipo_date"
    ipo_date = sa.Column(sa.DateTime(timezone=False))


class IPOPriceRange(AssetInfo):
    __tablename__ = "ipo_price"
    price_range_low = sa.Column(sa.Float, sa.CheckConstraint("price_range_low > 0"))
    price_range_high = sa.Column(
        sa.Float, sa.CheckConstraint("price_range_high > price_range_low")
    )


class DelistingDate(AssetInfo):
    __tablename__ = "delisting_date"
    delisting_date = sa.Column(sa.DateTime(timezone=False))


class Status(AssetInfo):
    __tablename__ = "status"
    status = sa.Column(sa.String)


class Currency(AssetInfo):
    __tablename__ = "currency"
    currency = sa.Column(sa.String)


class CurrencySymbol(AssetInfo):
    __tablename__ = "currency_symbol"
    currency_symbol = sa.Column(sa.String)


class Summary(AssetInfo):
    __tablename__ = "summary"
    summary = sa.Column(sa.String)


class Sector(AssetInfo):
    __tablename__ = "sector"
    sector = sa.Column(sa.String)


class Industry(AssetInfo):
    __tablename__ = "industry"
    industry = sa.Column(sa.String)


class MarketCategory(AssetInfo):
    __tablename__ = "market_category"
    market_category = sa.Column(sa.String)


class TestIssue(AssetInfo):
    __tablename__ = "test_issue"


class Office(AssetInfo):
    __tablename__ = "office"
    office = sa.Column(sa.String)


class FiscalYearEnd(AssetInfo):
    __tablename__ = "fiscal_year_end"
    fiscal_year_end = sa.Column(sa.String)


class FullTimeEmployees(AssetInfo):
    __tablename__ = "full_time_employees"
    full_time_employees = sa.Column(sa.String)


class CompanyOfficers(AssetInfo):
    __tablename__ = "company_officers"
    company_officers = sa.Column(sa.String)


class Logo(AssetInfo):
    __tablename__ = "logo"
    logo = sa.Column(sa.String)


class FIGI(AssetInfo):
    __tablename__ = "figi"
    figi = sa.Column(sa.String)


class ShareClassFIGI(AssetInfo):
    __tablename__ = "share_class_figi"
    share_class_figi = sa.Column(sa.String)


class Mic(AssetInfo):
    __tablename__ = "mic"
    mic = sa.Column(sa.String)


##### LOCATION #####


class Country(AssetInfo):
    __tablename__ = "country"
    country = sa.Column(sa.String)


class State(AssetInfo):
    __tablename__ = "state"
    state = sa.Column(sa.String)


class StateOfInc(AssetInfo):
    __tablename__ = "state_of_inc"
    state_of_inc = sa.Column(sa.String)


class City(AssetInfo):
    __tablename__ = "city"
    city = sa.Column(sa.String)


class ZipCode(AssetInfo):
    __tablename__ = "zip_code"
    zip_code = sa.Column(sa.String)


class BusinessAddress(AssetInfo):
    __tablename__ = "business_address"
    business_address = sa.Column(sa.String)


class MailingAddress(AssetInfo):
    __tablename__ = "mailing_address"
    mailing_address = sa.Column(sa.String)


##### Contact #####


class Phone(AssetInfo):
    __tablename__ = "phone"
    phone = sa.Column(sa.String)


class Fax(AssetInfo):
    __tablename__ = "fax"
    fax = sa.Column(sa.String)
