import asyncio
import signal
from itertools import cycle
from logging import Logger
from multiprocessing import Event, Pipe, Process
from multiprocessing.connection import Connection
from pathlib import Path
from typing import Any, Callable, List

from ready_logger import logger as default_logger


class DBLoader(Process):
    def __init__(
        self,
        data_pipe: Connection,
        shutdown_event: Event,
        load_cb: Callable,
        init_cb: Callable = None,
        logger: Logger = None,
    ):
        self.data_pipe = data_pipe
        self.shutdown_event = shutdown_event
        self.init_cb = init_cb
        self.load_cb = load_cb
        self.logger = logger or default_logger
        Process.__init__(self)

    def run(self):
        self.loop = asyncio.new_event_loop()
        self.loop.run_until_complete(self._run_async())

    async def _run_async(self):
        if self.init_cb:
            await self.init_cb()
        while True:
            data = self.data_pipe.recv()
            self.logger.info(
                f"Received batch of {len(data)} {type(data)}, e1: {type(data[0])}"
            )
            await self.load_cb(data)
            if self.shutdown_event.is_set():
                self.loop.stop()


async def cycle_fetch(
    fetch_cb,
    load_cb: Callable,
    batch_size: int,
    callback_cyclic_args: List[Any],
    logger=None,
    loop=None,
    **kwargs,
):
    logger = logger or default_logger
    shutdown_event = Event()
    data_sender, data_receiver = Pipe()
    db_loader = DBLoader(data_receiver, shutdown_event, load_cb, **kwargs)
    db_loader.start()

    batch = []

    def handler(signum, frame):
        logger.info(f"Caught signal {signum}. Shutting down.")
        shutdown_event.set()
        data_sender.send(batch)
        db_loader.join(60 * 60)
        if db_loader.is_alive():
            db_loader.terminate()
        data_sender.close()
        data_receiver.close()

    for s in (signal.SIGINT, signal.SIGTERM, signal.SIGHUP):
        signal.signal(s, handler)

    _cycle = cycle(callback_cyclic_args)

    async def run():
        while True:
            for _ in range(batch_size):
                if shutdown_event.is_set():
                    return
                arg = next(_cycle)
                resp = await fetch_cb(arg)
                batch.append(resp)
            data_sender.send(batch)

    await run()


loader_file = Path(__file__).parent / "loader.txt"


async def process_slow(data):
    await asyncio.sleep(2)
    with loader_file.open(mode="a+") as o:
        o.write(f"{data}\n")


writer_file = Path(__file__).parent / "writer.txt"


async def write_fast(arg):
    with writer_file.open(mode="a+") as o:
        o.write(f"{arg}\n")
    return arg


async def main():
    await cycle_fetch(
        fetch_cb=write_fast,
        load_cb=process_slow,
        batch_size=10,
        callback_cyclic_args=list(range(20)),
    )


if __name__ == "__main__":
    asyncio.run(main())
