import random
from datetime import datetime
from typing import List, Optional, Union

import sqlalchemy as sa
from database_utils.upsert_sources import SourcesUpserter, Upserter
from database_utils.utils import to_table
from ready_logger import logger
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.asyncio.engine import AsyncEngine
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlalchemy.sql import Select
from sqlalchemy.sql.elements import BooleanClauseList

from fin_data import asset_info as ai
from fin_data.sources.td_ameritrade.options.tables import (
    td_option_contracts,
    td_option_quotes,
    td_option_underlyings,
)


def symbols_least_recent_order_query(
    symbol_column: sa.Column, time_column: sa.Column
) -> Select:
    # TODO does this work with joined tables?
    """

    Args:
        symbol_column (sa.Column): [description]
        time_column (sa.Column): [description]

    Returns:
        Select: [description]
    """
    query = (
        sa.select(sa.column("symbol"))
        .select_from(
            sa.select(symbol_column, time_column)
            .distinct(symbol_column)
            .order_by(symbol_column, time_column.desc())
            .subquery()
        )
        .order_by(time_column.desc())
        .compile(dialect=postgresql.dialect())
    )
    return query


async def all_symbols_priority_order(
    engine: AsyncEngine,
    processed_symbols_column: sa.Column,
    processed_symbols_time_column: sa.Column,
    symbols_query: Select,
):
    async with engine.begin() as conn:
        processed_symbols = await conn.execute(
            symbols_least_recent_order_query(
                processed_symbols_column, processed_symbols_time_column
            )
        ).scalars()
        processed_symbols = list(processed_symbols)
        symbols_query = symbols_query.where(
            sa.columns("asset_id").notin_(processed_symbols)
        )
        new_symbols = list(conn.execute(symbols_query).scalars())
    random.shuffle(new_symbols)
    # make sure unprocessed symbols run first, then run in order of least recently processed.
    logger.info(
        f"Found {len(processed_symbols)} processed symbols, {len(new_symbols)} new symbols."
    )
    return new_symbols + processed_symbols


async def find_new_entities_query(
    sa_obj: Union[sa.Table, DeclarativeMeta],
    since_time: datetime,
    sources: Optional[Union[List[str], str]] = None,
) -> Select:
    """Find entities that were recently inserted added to the database.

    Args:
        sa_obj (Union[sa.Table, DeclarativeMeta]): The entity to check for new instances of.
        since_time (Union[datetime, str]): Oldest time that should be considered 'new'.
        count (bool, optional): Return number of entities instead of entity data rows. Defaults to False.
        source (Optional[str], optional): Restrict the query to a single source. Defaults to None.

    Returns:
        Select: the constructed query.
    """
    table = to_table(sa_obj)
    src_table = SourcesUpserter.sources_table(table)
    if isinstance(sources, str):
        sources = [sources]
    elif sources is None:
        sources = []
    query = (
        sa.select(sa_obj)
        .select_from(table)
        .join(src_table, src_table.c.data_id == table.c.data_id)
        .where(src_table.c.created <= since_time)
    )
    if sources:
        query = query.where(src_table.c.source == sources)
    return query


def most_recent_quotes():
    # find most recent quote time for each unique symbol.
    last_quote_times = (
        sa.select(td_option_quotes.c.symbol, td_option_quotes.c.quote_time)
        .distinct(td_option_quotes.c.symbol)
        .order_by(td_option_quotes.c.symbol, td_option_quotes.c.quote_time.asc())
        .subquery()
    )
    # self-join to all rows that have the symbol and it's most recent quote time.
    # compile with PostgreSQL dialect to generate the SELECT DISTINCT ON query.
    query = (
        sa.select(td_option_quotes)
        .select_from(
            last_quote_times.outerjoin(
                td_option_quotes,
                sa.and_(
                    last_quote_times.c.quote_time == td_option_quotes.c.quote_time,
                    last_quote_times.c.symbol == td_option_quotes.c.symbol,
                ),
            )
        )
        .compile(dialect=postgresql.dialect())
    )
    return query

def get_options(quotes_ignore_columns: List[str] = ['underlying_quote_id'],
                underlyings_ignore_columns: Union[List[str], bool] = ['id'],
                contracts_ignore_columns: Union[List[str], bool] = ['symbol', 'underlying_symbol'],
                not_expired: bool = True,
                oldest_quote_time: Optional[Union[str, datetime]] = None,
                underlying_symbols: Optional[List[str]] = None,
                where_conditions: Optional[BooleanClauseList] = None):
    to_select = []

    def add_selected_columns(table: sa.Table, ignore_columns: List[str] = [], column_name_prefix: Optional[str] = None):
        columns = [c for c in table.columns if c.name not in ignore_columns]
        if column_name_prefix:
            columns = [c.label(f'{column_name_prefix}{c.name}') for c in columns]
        to_select.extend(columns)

    # table(s) that data should be selected from.
    select_from = td_option_quotes

    add_selected_columns(td_option_quotes, ignore_columns=quotes_ignore_columns or [])
    if underlyings_ignore_columns is not True:
        add_selected_columns(td_option_underlyings, ignore_columns=underlyings_ignore_columns or [], column_name_prefix='underlying_')
    if underlyings_ignore_columns is not True or underlying_symbols is not None:
        select_from = select_from.outerjoin(td_option_underlyings, td_option_underlyings.c.id == td_option_quotes.c.underlying_quote_id)
    if contracts_ignore_columns is not True:
        add_selected_columns(td_option_contracts, ignore_columns=contracts_ignore_columns or [])
    if contracts_ignore_columns is not True or not_expired:
        select_from = select_from.outerjoin(td_option_contracts, td_option_contracts.c.symbol == td_option_quotes.c.symbol)

    # start building query.
    query = sa.select(to_select)

    # add filters.
    if isinstance(oldest_quote_time, datetime):
        query = query.where(td_option_quotes.c.quote_time > oldest_quote_time)
    elif oldest_quote_time == 'latest':
        # find most recent quote time for each unique symbol.
        last_quote_times = (
            sa.select(td_option_quotes.c.symbol, td_option_quotes.c.quote_time)
            .distinct(td_option_quotes.c.symbol)
            .order_by(td_option_quotes.c.symbol, td_option_quotes.c.quote_time.desc())
            .subquery()
        )
        # self-join to all rows that have the symbol and it's most recent quote time.
        # compile with PostgreSQL dialect to generate the SELECT DISTINCT ON query.
        select_from = last_quote_times.outerjoin(select_from, sa.and_(last_quote_times.c.quote_time == td_option_quotes.c.quote_time, last_quote_times.c.symbol == td_option_quotes.c.symbol))

    if where_conditions is not None:
        query = query.where(where_conditions)
    if not_expired:
        query = query.where(td_option_contracts.c.expiration_date > datetime.utcnow())
    if underlying_symbols is not None:
        if not isinstance(underlying_symbols, (list,tuple,set)):
            underlying_symbols = [underlying_symbols]
        query = query.where(td_option_contracts.c.underlying_symbol.in_(underlying_symbols))

    return query.select_from(select_from).compile(dialect=postgresql.dialect())


REGISTRY = {
    "all_symbols_priority_order": all_symbols_priority_order,
}
