from datetime import datetime
from functools import partial

from database_utils.upsert_sources import SourcesUpserter
from fin_data import asset_info as ai
from fin_data.sources.alphavantage.utils import fetch, get_api_key, load_csv_with_header
from fin_data.utils import load_csv_with_header
from ready_logger import get_logger

logger = get_logger(name="alphavantage_ipo")

# create the listing status loader.
UpserterT = partial(
    SourcesUpserter.create,
    source="alphavantage",
    logger=logger,
)


async def load_ipo_calender():
    """(CSV) This API returns a list of IPOs expected in the next 3 months."""
    api_key = get_api_key()
    url = f"https://www.alphavantage.co/query?function=IPO_CALENDAR&apikey={api_key}"

    response_text = await fetch(url)
    # load to CSV.
    rows = load_csv_with_header(response_text)

    row_ids = [
        {"asset_id": row["symbol"], "asset_id_type": ai.CompanyIDType.SYMBOL}
        for row in rows
    ]

    # add name.
    name_loader = await UpserterT(sa_obj=ai.Name)
    await name_loader.upsert(
        [
            {
                "name": row["name"],
                **row_id,
            }
            for row, row_id in zip(rows, row_ids)
        ]
    )

    # add exchange.
    exchange_loader = await UpserterT(sa_obj=ai.Exchange)
    await exchange_loader.upsert(
        [
            {
                "exchange": row["exchange"],
                **row_id,
            }
            for row, row_id in zip(rows, row_ids)
        ]
    )

    # add ipo date.
    ipo_date_loader = await UpserterT(sa_obj=ai.IPODate)
    await ipo_date_loader.upsert(
        [
            {
                "ipo_date": datetime.fromisoformat(row["ipoDate"]),
                **row_id,
            }
            for row, row_id in zip(rows, row_ids)
        ]
    )

    # add ipo price range.
    ipo_price_loader = await UpserterT(sa_obj=ai.IPOPriceRange)
    await ipo_price_loader.upsert(
        [
            {
                "price_range_low": row["priceRangeLow"],
                "price_range_high": row["priceRangeHigh"],
                **row_id,
            }
            for row, row_id in zip(rows, row_ids)
        ]
    )

    currency_loader = await UpserterT(sa_obj=ai.Currency)
    await currency_loader.upsert(
        [
            {
                "currency": row["currency"],
                **row_id,
            }
            for row, row_id in zip(rows, row_ids)
        ]
    )
