import asyncio
from datetime import datetime
from functools import partial

from database_utils.upsert_sources import SourcesUpserter
from fin_data import asset_info as ai
from fin_data.sources.alphavantage.utils import (
    SOURCE_NAME,
    fetch,
    get_api_key,
    load_csv_with_header,
)
from ready_logger import get_logger

logger = get_logger("alphavantage-listing-status")

UpserterT = partial(
    SourcesUpserter.create,
    source=SOURCE_NAME,
    logger=logger,
)


async def alphavantage_listing_status():
    """(CSV) This API returns ay list of active or delisted US stocks and ETFs,
    either as of the latest trading day or at a specific time in history.
    The endpoint is positioned to facilitate equity research on asset lifecycle and survivorship.
    Optional: date
    If no date is set, the API endpoint will return a list of active or delisted symbols as of the latest trading day.
    If a date is set, the API endpoint will "travel back" in time and return a list of active or delisted symbols on that particular date in history.
    Any YYYY-MM-DD date later than 2010-01-01 is supported. For example, date=2013-08-03
    Optional: state
    By default, state=active and the API will return a list of actively traded stocks and ETFs.
    Set state=delisted to query a list of delisted assets.
    """
    api_key = get_api_key()

    # fetch the file.
    url = f"https://www.alphavantage.co/query?function=LISTING_STATUS&apikey={api_key}"
    response_text = await fetch(url)

    # load to CSV.
    rows = load_csv_with_header(response_text)

    # load names.
    name_loader = await UpserterT(sa_obj=ai.Name)
    await name_loader.upsert(
        [
            {
                "name": row["name"],
                "asset_id": row["symbol"],
                "asset_id_type": ai.CompanyIDType.SYMBOL,
            }
            for row in rows
        ]
    )

    # load exchanges.
    exchange_loader = await UpserterT(sa_obj=ai.Exchange)
    await exchange_loader.upsert(
        [
            {
                "exchange": row["exchange"],
                "asset_id": row["symbol"],
                "asset_id_type": ai.CompanyIDType.SYMBOL,
            }
            for row in rows
        ]
    )

    # load delisting date.
    delisting_loader = await UpserterT(sa_obj=ai.DelistingDate)
    await delisting_loader.upsert(
        [
            {
                "delisting_date": datetime.fromisoformat(row["delistingDate"]),
                "asset_id": row["symbol"],
                "asset_id_type": ai.CompanyIDType.SYMBOL,
            }
            for row in rows
            if row["delistingDate"] != "null"
        ]
    )

    # add ipo date.
    ipo_date_loader = await UpserterT(sa_obj=ai.IPODate)
    await ipo_date_loader.upsert(
        [
            {
                "ipo_date": datetime.fromisoformat(row["ipoDate"]),
                "asset_id": row["symbol"],
                "asset_id_type": ai.CompanyIDType.SYMBOL,
            }
            for row in rows
        ]
    )

    type_loader = await UpserterT(sa_obj=ai.Type)
    await type_loader.upsert(
        [
            {
                "type": row["assetType"],
                "asset_id": row["symbol"],
                "asset_id_type": ai.CompanyIDType.SYMBOL,
            }
            for row in rows
        ]
    )


if __name__ == "__main__":
    from dotenv import load_dotenv

    load_dotenv("/home/dan/danklabs/.localenv")

    asyncio.run(alphavantage_listing_status())
