import asyncio
import csv
import os
from io import StringIO
from typing import Any, Dict, List

import requests
from host_throttle import RateLimiter
from ready_logger import logger

ALPHAVANTAGE_API_KEY_ENV_VAR = "ALPHAVANTAGE_API_KEY"
SOURCE_NAME = "alphavantage"

request_count = 0
rate_limiter = RateLimiter.from_constant_rate(n_requests=5, seconds=60)


async def fetch(url: str, return_json: bool = False) -> str:
    """[summary]

    Args:
        url ([type]): [description]
        json (bool, optional): [description]. Defaults to False.

    Raises:
        RuntimeError: [description]

    Returns:
        str: [description]
    """
    global request_count
    # 500 requests per day, 5 request per minute.
    if request_count > 500:
        raise RuntimeError(f"Exceeded max daily requests (500)")
    await rate_limiter.maybe_sleep(url)
    resp = requests.get(url)
    request_count += 1
    logger.info(f"[{resp.status_code}] {url}")
    if "Our standard API call frequency is" in resp.text:
        await asyncio.sleep(60 * 5)
        return await fetch(url, return_json)
    if resp.status_code == 200:
        if return_json:
            return resp.json()
        return resp.text
    if resp.status_code == 429:
        await asyncio.sleep(20)
        return await fetch(url)
    raise RuntimeError(f"Error fetching: {url}. Status code {resp.status_code}")


def get_api_key() -> str:
    if (api_key := os.getenv(ALPHAVANTAGE_API_KEY_ENV_VAR)) is None:
        raise RuntimeError(
            f"Environmental variable {ALPHAVANTAGE_API_KEY_ENV_VAR} is not set."
        )
    return api_key


def load_csv_with_header(file_text: str) -> List[Dict[str, Any]]:
    fo = StringIO(file_text)
    header = [c.strip() for c in next(csv.reader(fo))]
    rows = list(csv.DictReader(fo, fieldnames=header))
    logger.info(f"Loaded {len(rows)} rows from CSV.")
    return rows
