import os
from functools import partial
from typing import Dict, List

from aiohttp import ClientSession
from fin_data import asset_info as ai
from fin_data.utils import create_http_client, extract_column_name
from database_utils.upsert_sources import SourcesUpserter
from host_throttle import RateLimiter
from ready_logger import get_logger

# rate limit: 60 API calls/minute
rate_limiter = RateLimiter.from_rate_limit(n_requests=1, seconds=1)

logger = get_logger("Finnhub")
UpserterT = partial(SourcesUpserter.create, source="finnhub", logger=logger)


async def get_json(url: str, client: ClientSession):
    async with client.get(url) as resp:
        logger.info("[%i] %s", resp.status, url)
    if resp.status == 200:
        return resp.json()
    if resp.status == 429:
        rate_limiter.pause(url, 20)
        return await get_json(url, client)


async def exchange_symbols(exchanges: List[str] = ["US"]) -> List[Dict[str, str]]:
    """List supported stocks. We use the following symbology to identify stocks on Finnhub Exchange_Ticker. Exchange_Code.
    A list of supported exchange codes can be found here: https://docs.google.com/spreadsheets/d/1I3pBxjfXB056-g_JYf_6o3Rns3BV2kMGG1nCatb91ls/edit#gid=0
    A list of supported CFD Indices can be found here: https://docs.google.com/spreadsheets/d/1BAbIXBgl405fj0oHeEyRFEu8mW4QD1PhvtaBATLoR14/edit#gid=0
    """
    column_value_upserters = {
        field: (await UpserterT.create(entity), extract_column_name(entity))
        for field, entity in {
            "currency": ai.Currency,
            "description": ai.Name,
            "figi": ai.Figi,
            "shareClassFIGI": ai.ShareClassFIGI,
            "mic": ai.Mic,
            "name": ai.Name,
            "type": ai.Type,
            "exchange": ai.Exchange,
        }.items()
    }
    async with create_http_client() as client:
        for exchange in exchanges:
            data = await get_json(
                f"https://finnhub.io/api/v1/stock/symbol?exchange={exchange}&token={os.getenv('FINNHUB_API_KEY')}",
                client,
            )
            data["exchange"] = exchange
            for asset_id in {data["symbol"], data["displaySymbol"]}:
                asset_id_data = {
                    "asset_id": asset_id,
                    "asset_id_type": ai.CompanyIDType.SYMBOL,
                }
                for field, (upserter, column_name) in column_value_upserters.items():
                    if field in data:
                        await upserter.upsert(
                            {**asset_id_data, column_name: data[field]}
                        )


async def company_info_batch(symbols: List[str]):
    column_value_upserters = {
        field: (await UpserterT.create(entity), extract_column_name(entity))
        for field, entity in {
            "country": ai.Country,
            "exchange": ai.Exchange,
            "finnhubIndustry": ai.Industry,
            "logo": ai.Logo,
            "name": ai.Name,
            "phone": ai.Phone,
            "weburl": ai.Website,
            "ipo": ai.IPODate,
        }.items()
    }
    async with create_http_client() as client:
        for symbol in symbols:
            data = get_json(
                f"https://finnhub.io/api/v1/stock/profile2?symbol={symbol}&token={os.getenv('FINNHUB_API_KEY')}",
                client,
            )
            if "symbol" not in data:
                data["symbol"] = symbol
                logger.info("Warning. Could not get data for symbol: %s", symbol)
            for asset_id in {data["symbol"], data["ticker"]}:
                asset_id_data = {
                    "asset_id": asset_id,
                    "asset_id_type": ai.CompanyIDType.SYMBOL,
                }
                for field, (upserter, column_name) in column_value_upserters.items():
                    if field in data:
                        await upserter.upsert(
                            {**asset_id_data, column_name: data[field]}
                        )


async def peers(symbol: str):
    """Get company peers. Return a list of peers in the same country and GICS sub-industry"""
    return await get_json(
        f"https://finnhub.io/api/v1/stock/peers?symbol={symbol}&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def indices_constituents(symbol: str):
    """Get a list of index's constituents.
    Currently support ^GSPC (S&P 500), ^NDX (Nasdaq 100), ^DJI (Dow Jones)
    """
    return await get_json(
        f"https://finnhub.io/api/v1/index/constituents?symbol={symbol}&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def etf_industry_exposure(symbol: str):
    """Get ETF industry exposure data"""
    return await get_json(
        f"https://finnhub.io/api/v1/etf/sector?symbol={symbol}&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def etf_country_exposure(symbol: str):
    """Get ETF country exposure data."""
    return await get_json(
        f"https://finnhub.io/api/v1/etf/country?symbol={symbol}&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def financials(symbol: str):
    """Get financials as reported."""
    return await get_json(
        f"https://finnhub.io/api/v1/stock/financials-reported?symbol={symbol}&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def sec_filings(symbol: str):
    """List company's filing. Limit to 250 documents at a time.
    Symbol. Leave symbol,cik and accessNumber empty to list latest filings.
    """
    return await get_json(
        f"https://finnhub.io/api/v1/stock/filings?symbol={symbol}&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def basic_financials(symbol: str):
    """Get company basic financials such as margin, P/E ratio, 52-week high/low etc."""
    return await get_json(
        f"https://finnhub.io/api/v1/stock/metric?symbol={symbol}&metric=all&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def economic_calender():
    """Get recent and upcoming economic releases.
    Historical events and surprises are available for Enterprise clients.
    """
    return await get_json(
        f"https://finnhub.io/api/v1/calendar/economic?token={os.getenv('FINNHUB_API_KEY')}"
    )


async def latest_market_news():
    return await get_json(
        f"https://finnhub.io/api/v1/news?category=general&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def company_news(symbol: str):
    """List latest company news by symbol. This endpoint is only available for North American companies.
    Free Tier: 1 year of historical news and new updates
    """
    return await get_json(
        f"https://finnhub.io/api/v1/company-news?symbol={symbol}&from=2020-04-30&to=2020-05-01&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def ipo_calender():
    return await get_json(
        f"https://finnhub.io/api/v1/calendar/ipo?from=2020-01-01&to=2020-04-30&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def recommendation_trends(symbol: str):
    return await get_json(
        f"https://finnhub.io/api/v1/stock/recommendation?symbol={symbol}&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def earnings_suprises(symbol: str):
    """Get company historical quarterly earnings surprise going back to 2000"""
    return await get_json(
        f"https://finnhub.io/api/v1/stock/earnings?symbol={symbol}&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def news_sentiment(symbol: str):
    """Get company's news sentiment and statistics. This endpoint is only available for US companies."""
    return await get_json(
        f"https://finnhub.io/api/v1/news-sentiment?symbol={symbol}&token={os.getenv('FINNHUB_API_KEY')}"
    )


async def earnings():
    """Get historical and coming earnings release.
    EPS and Revenue in this endpoint are non-GAAP, which means they are adjusted to exclude some one-time or unusual items.
    This is the same data investors usually react to and talked about on the media.
    Estimates are sourced from both sell-side and buy-side analysts.
    from optional
    From date: 2020-03-15.

    to optional
    To date: 2020-03-16.

    symbol optional
    Filter by symbol: AAPL.

    internationaloptional
    Set to true to include international markets. Default value is false
    """
    return await get_json(
        f"https://finnhub.io/api/v1/calendar/earnings?&token={os.getenv('FINNHUB_API_KEY')}"
    )
