import json
import os
from datetime import datetime
from pprint import pformat

import sqlalchemy as sa
from aiohttp import ClientSession
from ready_logger import get_logger
from sqlalchemy.dialects.postgresql import ARRAY, JSON

metadata = sa.MetaData()

place_search_table = sa.Table(
    "google_place_search",
    metadata,
    sa.Column("query", sa.String, primary_key=True),
    sa.Column("place_ids", ARRAY(sa.String, dimensions=1), primary_key=True),
    sa.Column("created", sa.DateTime(timezone=False), default=datetime.utcnow),
    sa.Column("last_checked", sa.DateTime(timezone=False), default=datetime.utcnow),
)

place_detail_table = sa.Table(
    "google_place_detail",
    metadata,
    sa.Column("data_id", sa.String, primary_key=True),
    sa.Column("place_id", sa.String),
    sa.Column("business_status", sa.String),
    sa.Column("business_address", sa.String),
    sa.Column("html_business_address", sa.String),
    sa.Column("logo", sa.String),
    sa.Column("phone", sa.String),
    sa.Column("name", sa.String),
    sa.Column("rating", sa.Float),
    sa.Column("maps_url", sa.String),
    sa.Column("user_ratings_total", sa.Integer),
    sa.Column("photos", JSON()),
    sa.Column("reviews", JSON()),
    sa.Column("opening_hours", ARRAY(sa.String, dimensions=1)),
    sa.Column("address_components", JSON()),
    sa.Column("website", sa.String),
    sa.Column(
        "created", sa.DateTime(timezone=False), default=datetime.utcnow, index=True
    ),
    sa.Column(
        "last_checked", sa.DateTime(timezone=False), default=datetime.utcnow, index=True
    ),
)


async def place_search(query: str, client: ClientSession):
    params = {
        "input": query,
        "inputtype": "textquery",
        "key": os.getenv("GOOGLE_PLACES_API_KEY"),
    }
    endpoint = "https://maps.googleapis.com/maps/api/place/findplacefromtext/json"
    async with client.get(endpoint, params=params) as resp:
        data = resp.json()
        logger.info(f"[{resp.status_code}] {endpoint}: {pformat(data)}")
    return [c["place_id"] for c in data.get("candidates", [])]


field_map = {
    "business_status": "business_status",
    "formatted_address": "business_address",
    "adr_address": "html_business_address",
    "icon": "logo",
    "international_phone_number": "phone",
    "name": "name",
    "opening_hours": "opening_hours",
    "photos": "photos",
    "rating": "rating",
    "reviews": "reviews",
    "url": "maps_url",
    "user_ratings_total": "user_ratings_total",
    "website": "website",
    "address_components": "address_components",
    "place_id": "place_id",
}


def place_details(place_id: str):
    print(f"Searching place details: {place_id}")
    params = {
        "place_id": place_id,
        "key": os.getenv("GOOGLE_PLACES_API_KEY"),
    }
    endpoint = "https://maps.googleapis.com/maps/api/place/details/json?parameters"
    resp = requests.get(endpoint, params=params)
    data = resp.json()
    print(f"[{resp.status_code}] {endpoint}: {pformat(data)}")
    if result := data.get("result"):
        data = {field_map[k]: v for k, v in result.items() if k in field_map}
        if "reviews" in data:
            data["reviews"] = json.dumps(data["reviews"])
        if "photos" in data:
            data["photos"] = json.dumps(data["photos"])
        if "address_components" in data:
            data["address_components"] = json.dumps(data["address_components"])
        if "opening_hours" in data:
            data["opening_hours"] = data["opening_hours"]["weekday_text"]
        return data
