from utils import get_sql_conn, fetch, get_spider
from update_table import insert_or_update_rows

import aioredis
from urllib.parse import quote
from typing import List
from itertools import product
from pprint import pformat
from hashlib import md5
from pathlib import Path
from tqdm import tqdm
import argparse
from random import uniform
import logging
import asyncio
import json
import re

spider = redis_client = None


def get_search_id(search_params):
    return md5(json.dumps(search_params).encode()).hexdigest()


def get_industry_searches():
    industry_search_params = {
        "geography": {
            "United States": 103644278,
        },
        "industry": {
            # 'Computer Software': 4,
            # 'Banking': 41,
            "Financial Services": 43,
            "Investment Banking": 45,
            "Investment Management": 46,
        },
        "functions": {
            "Finance": 10,
            "Sales": 25,
            "Business Development": 4,
            "Operations": 18,
            "Information Technology": 13,
            "Engineering": 8,
            # 'Human Resources': 12,
            "Research": 24,
            "Marketing": 15,
            "Consulting": 6,
            "Community and Social Services": 5,
            "Education": 7,
            "Entrepreneurship": 9,
            "Product Management": 19,
        },
        "company_headcount": {
            "Self-employed": "A",
            "1-10": "B",
            "11-50": "C",
            "51-200": "D",
            "201-500": "E",
            "501-1000": "F",
            "1001-5000": "G",
            "5001-10,000": "H",
            "10,000+": "I",
        },
        "seniority_level": {
            "Owner": 10,
            "Partner": 9,
            "CXO": 8,
            "VP": 7,
            "Director": 6,
            "Manager": 5,
            "Senior": 4,
            "Entry": 3,
            "Training": 2,
            "Unpaid": 1,
        },
    }
    searches = []
    # loop through all permutations of search parameters.
    for search in product(
        *[
            [(k, kk, vv) for kk, vv in v.items()]
            for k, v in industry_search_params.items()
        ]
    ):
        params = {
            field: {"value": value, "url_param": url_param}
            for field, value, url_param in search
        }
        searches.append(
            {
                "params": params,
                "search_id": get_search_id(params),
                "search_url": "https://www.linkedin.com/sales/search/people?"
                f"companySize={params['company_headcount']['url_param']}"
                f"&functionIncluded={params['functions']['url_param']}"
                f"&geoIncluded={params['geography']['url_param']}"
                f"&industryIncluded={params['industry']['url_param']}"
                f"&seniorityIncluded={params['seniority_level']['url_param']}"
                "&doFetchHeroCard=false&logHistory=false&page=1",
            }
        )
    return searches


def get_company_search(company_name):
    params = {"company_name": company_name}
    return {
        "params": params,
        "search_id": get_search_id(params),
        "search_url": f"https://www.linkedin.com/sales/search/people?"
        f"companyIncluded={re.sub(r'%','%25',quote(company_name))}"
        "&companyTimeScope=CURRENT&doFetchHeroCard=false&logHistory=false&page=1",
    }


async def scroll_page(page):
    try:
        for i in range(1, 7):
            await page.evaluate(f"window.scrollTo(0, {i*500});")
            await asyncio.sleep(0.1)
    except Exception as e:
        logging.exception(f"Page ({page.url}) not scrollable. Error: {e}")


async def scrape_result_page_profile(prof_ele, page):
    "Extract profile info from search results page."
    profile = {}
    name_and_link = await prof_ele.xpath('.//*[@class="result-lockup__name"]/a')
    if len(name_and_link):
        profile["name"] = await page.evaluate(
            "(ele) => ele.innerText", name_and_link[0]
        )
        person_profile_link = await page.evaluate(
            "(ele) => ele.getAttribute('href')", name_and_link[0]
        )
        profile[
            "profile_url"
        ] = f"https://www.linkedin.com{person_profile_link.strip()}"
    title = await prof_ele.xpath('.//*[@class="result-lockup__highlight-keyword"]/span')
    if len(title):
        person_title = await page.evaluate("(ele) => ele.innerText", title[0])
        profile["title"] = person_title.strip()
    company = await prof_ele.xpath(
        './/*[@class="result-lockup__highlight-keyword"]//span/a'
    )
    if len(company):
        profile["company_url"] = await page.evaluate(
            "(ele) => ele.getAttribute('href')", company[0]
        )
        company_name = await company[0].xpath("./span")
        if len(company_name):
            company_name = await page.evaluate(
                "(ele) => ele.innerText", company_name[0]
            )
            profile["company_name"] = company_name.strip()
    location = await prof_ele.xpath('.//*[@class="result-lockup__misc-item"]')
    if len(location):
        location = await page.evaluate("(ele) => ele.innerText", location[0])
        profile["location"] = location.strip()
    logging.info(f"Extracted profile:\n{pformat(profile)}")
    return profile


async def run_search(search_params, page_limit: int = None):
    json_search_params = json.dumps(search_params)

    async def scrape_results_page(page, pg_num, search_params, profiles):
        logging.info(f"Scraping results page {pg_num}: {page.url}")
        await scroll_page(page)
        prof_eles = await page.xpath('//*[@class="result-lockup"]')
        for ele in tqdm(prof_eles):
            await ele.hover()
            prof = await scrape_result_page_profile(ele, page)
            prof["results_page_num"] = pg_num
            prof["search_params"] = json_search_params
            profiles.append(prof)
        # mark page number as complete.
        await redis_client.sadd(search_params["search_id"], pg_num)

    async def get_unprocessed_pages(last_page: int) -> List[int]:
        remaining_pages = []
        for pg_num in range(1, last_page + 1):
            is_proc = await redis_client.sismember(search_params["search_id"], pg_num)
            if not is_proc:
                remaining_pages.append(pg_num)
        logging.info(
            f"Found {last_page-len(remaining_pages)} processed pages, {len(remaining_pages)} remaining."
        )
        return remaining_pages

    logging.info(f"Starting search: {pformat(search_params)}")
    # get last page number.
    page = await fetch(spider, search_params["search_url"])
    pagination_eles = await page.xpath(
        '//button[contains(@aria-label,"Navigate to page")]'
    )
    if pagination_eles:
        last_page = await page.evaluate(
            '(ele) => ele.getAttribute("aria-label")', pagination_eles[-1]
        )
        last_page = int(last_page.replace("Navigate to page", "").strip())
        if page_limit is not None:
            last_page = min(page_limit, last_page)
    else:
        last_page = 1
    await spider.set_idle(page)
    logging.info(f"Found last page number: {last_page}")
    unprocessed_pages = await get_unprocessed_pages(last_page)
    profiles = []
    if len(unprocessed_pages) and unprocessed_pages[0] == 1:
        # already on page 1, so don't navigate.
        await scrape_results_page(
            page, unprocessed_pages.pop(0), search_params, profiles
        )
    for pg_num in tqdm(unprocessed_pages):
        pg_url = re.sub(r"page=\d{1,3}", f"page={pg_num}", page.url)
        await spider.set_idle(page)
        # navigate to page. get current url from page and set new page number.
        page = await fetch(spider, pg_url)
        await scrape_results_page(page, pg_num, search_params, profiles)
    await spider.set_idle(page)
    logging.info(
        f"Finished search {search_params['search_id']}: {pformat(search_params)}"
    )
    return profiles


async def queue_industry_searches():
    searches = [json.dumps(search) for search in get_industry_searches()]
    logging.info(f"Queueing {len(searches)} industry searches")
    await redis_client.sadd("linkedin_industry_searches", *searches)


async def run_all_searches(key, conn, page_limit: int = None):
    while True:
        data = await redis_client.spop(key, encoding="utf-8")
        if data is None:
            logging.info("Finished all searches.")
            return
        if key == "linkedin_industry_searches":
            search_params = json.loads(data)
        elif key == "linkedin_company_names":
            search_params = get_company_search(data)
        profiles = await run_search(search_params, page_limit)
        insert_or_update_rows(
            conn=conn,
            rows=profiles,
            table_name="linkedin_salesnav_search",
            prim_key_col="profile_url",
            update_columns=[
                "name",
                "profile_url",
                "title",
                "company_url",
                "company_name",
                "location",
                "results_page_num",
                "search_params",
            ],
        )


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-q", "--queue", help="Queue searches.", action="store_true")
    parser.add_argument(
        "-t",
        "--type",
        help="Type of search to run (industry or company)",
        type=str,
        required=True,
    )
    parser.add_argument("-l", "--page_limit", type=int)
    return parser.parse_args()


async def main():
    global redis_client, spider
    args = parse_args()
    redis_client = await aioredis.create_redis("redis://127.0.0.1:6379/0")
    spider = await get_spider()
    conn = get_sql_conn()
    page = await fetch(spider, "https://www.linkedin.com/uas/login")
    await asyncio.sleep(3)
    await spider.set_idle(page)
    # nav to homepace.
    _, page = await spider.get(
        "https://www.linkedin.com/sales/homepage", waitUntil=["load", "networkidle2"]
    )
    await asyncio.sleep(3)
    await spider.set_idle(page)
    # parse search type from args.
    if args.type.startswith("i"):
        if args.queue:
            await queue_industry_searches()
        await run_all_searches("linkedin_industry_searches", conn, args.page_limit)
    elif args.type.startswith("c"):
        await run_all_searches("linkedin_company_names", conn, args.page_limit)
    conn.close()
    redis_client.close()
    await spider.shutdown()


if __name__ == "__main__":
    asyncio.run(main())
