from utils import get_sql_conn

from psycopg2.extras import RealDictCursor
from typing import List, Dict, Union
from datetime import datetime, timedelta
from tqdm import tqdm


def clean_row(row):
    return {k: v if v not in ("null", "") else None for k, v in row.items()}


def insert_or_update_rows(
    rows: List[Dict[str, str]],
    table_name: str,
    prim_key_col: Union[str, List[str]],
    update_columns: Union[str, List[str]] = [],
    conn=None,
):
    def try_execute(cur, statement, row):
        try:
            cur.execute(statement, row)
        except Exception as e:
            raise Exception(f"ERROR: {e}. \n\nROW: {row}. \n\nSTATEMENT: {statement}\n")

    _conn = get_sql_conn() if conn is None else conn
    if not isinstance(rows, (list, tuple)):
        rows = [rows]
    rows = [clean_row(row) for row in rows]
    prim_key_col = (
        {prim_key_col} if isinstance(prim_key_col, str) else set(prim_key_col)
    )
    update_columns = (
        {update_columns} if isinstance(update_columns, str) else set(update_columns)
    )
    update_columns.difference_update(prim_key_col)
    # WHERE condition for querying on primary key.
    where_condition = " AND ".join([f"{k}=%({k})s" for k in prim_key_col])
    # if UPDATEing columns, SELECT the current column values so we can log differences between existing and new row.
    select_arg = ",".join(update_columns) if len(update_columns) else "1"
    select_statement = f"SELECT {select_arg} FROM {table_name} WHERE {where_condition}"
    updated_rows, new_rows = [], []
    with _conn.cursor(cursor_factory=RealDictCursor) as cur:
        for row in tqdm(rows):
            # check if row with this key currently exists.
            try_execute(cur, select_statement, row)
            current_row = list(cur.fetchall())
            if len(current_row):
                if len(current_row) > 1:
                    raise ValueError(
                        f"""Select statement ({select_statement}) returned {len(current_row)} rows.
                            Column(s) passed to prim_key_col parameter ({prim_key_col}) should have unique contraint.
                            Keys used: {{{k: row.get(k) for k in prim_key_col}}}"""
                    )
                # row exists. update time that row was last seen in the file.
                to_set = {"last_seen=now()"}
                # check if any columns should be updated.
                if len(update_columns):
                    # remove row dict from list
                    current_row = current_row.pop()
                    # check all keys to see if any values have changes.
                    keys = {*current_row.keys(), *row.keys()}
                    updated_cols = {
                        k: row.get(k) for k in keys if current_row.get(k) != row.get(k)
                    }
                    if len(updated_cols):
                        # format for update using dict.
                        to_set.update([f"{k}=%({k})s" for k in updated_cols.keys()])
                        # record time of update.
                        to_set.add("updated=now()")
                        # increment update count.
                        to_set.add("updates=updates+1")
                        # save data to be returned.
                        updated_rows.append(
                            {
                                "row_key": {k: row.get(k) for k in prim_key_col},
                                "updates": updated_cols,
                            }
                        )
                # perform update on row.
                try_execute(
                    cur,
                    f"UPDATE {table_name} SET {','.join(list(to_set))} WHERE {where_condition}",
                    row,
                )
            else:
                # no row with this key currently exists, so insert a new row.
                try_execute(
                    cur,
                    f"INSERT INTO {table_name} ({','.join(row.keys())}) VALUES ({','.join([f'%({k})s' for k in row.keys()])})",
                    row,
                )
                new_rows.append(row)
    # commit all inserts/updates to table.
    _conn.commit()
    if conn is None:
        _conn.close()
    print(
        f"[table {table_name}] Updated {len(updated_rows)} existing rows. Inserted {len(new_rows)} new rows."
    )
    return updated_rows, new_rows


def check_newly_delisted(
    conn, table_name: str, start_time: datetime
) -> List[Dict[str, str]]:
    with conn.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
        # find all row that were not updated in the current update and have not been marked as delisted.
        cur.execute(
            f"SELECT * FROM {table_name} WHERE last_seen < %s AND delisted IS NULL",
            (start_time,),
        )
        newly_delisted = list(cur.fetchall())
        # mark as delisted.
        cur.execute(
            f"UPDATE {table_name} SET delisted = %s WHERE last_seen < %s AND delisted IS NULL",
            (start_time, start_time),
        )
    conn.commit()
    print(f"[table {table_name}] Found {len(newly_delisted)} newly delisted.")
    return newly_delisted


def update_table(
    rows: List[Dict[str, str]],
    table_name: str,
    prim_key_col: Union[str, List[str]],
    update_columns: List[str] = [],
    conn=None,
):
    _conn = get_sql_conn() if conn is None else conn
    start = datetime.now()
    updated_rows, new_rows = insert_or_update_rows(
        _conn, rows, table_name, prim_key_col, update_columns
    )
    newly_delisted = check_newly_delisted(_conn, table_name, start)
    if conn is None:
        _conn.close()
    return new_rows, updated_rows, newly_delisted
