from functools import partial

from database_utils.upsert_sources import SourcesUpserter
from fin_data import asset_info as ai
from fin_data.asset_info import CompanyIDType
from fin_data.sources.nasdaq.utils import downloaded_file, load_rows
from ready_logger import get_logger

market_category = {
    "Q": "NASDAQ Global Select MarketSM",
    "G": "NASDAQ Global MarketSM",
    "S": "NASDAQ Capital Market",
}

financial_status = {
    "D": "Deficient: Issuer Failed to Meet NASDAQ Continued Listing Requirements",
    "E": "Delinquent: Issuer Missed Regulatory Filing Deadline",
    "Q": "Bankrupt: Issuer Has Filed for Bankruptcy",
    "N": "Normal (Default): Issuer Is NOT Deficient, Delinquent, or Bankrupt.",
    "G": "Deficient and Bankrupt",
    "H": "Deficient and Delinquent",
    "J": "Delinquent and Bankrupt",
    "K": "Deficient, Delinquent, and Bankrupt",
}

logger = get_logger(name="nasdaqlisted")

UpserterT = partial(
    SourcesUpserter.create,
    source="nasdaq nasdaqlisted.txt",
    logger=logger,
)


async def load_nasdaqlisted() -> None:
    with downloaded_file("nasdaqlisted.txt") as file:
        rows, timestamp = load_rows(file)
        logger.info(f"Loaded {len(rows)} rows from file. File update time {timestamp}")

        name_loader = await UpserterT(sa_obj=ai.Name)
        await name_loader.upsert(
            [
                {
                    "name": row["Security Name"],
                    "name_type": ai.NameType.ASSET,
                    "asset_id": row["Symbol"],
                    "asset_id_type": CompanyIDType.SYMBOL,
                }
                for row in rows
            ]
        )

        type_loader = await UpserterT(sa_obj=ai.Type)
        await type_loader.upsert(
            [
                {
                    "asset_id": row["Symbol"],
                    "asset_id_type": CompanyIDType.SYMBOL,
                    "type": "ETF",
                }
                for row in rows
                if row["ETF"] == "Y"
            ]
        )

        exchange_loader = await UpserterT(sa_obj=ai.Exchange)
        await exchange_loader.upsert(
            [
                {
                    "exchange": market_category.get(
                        row["Market Category"], row["Market Category"]
                    ),
                    "asset_id": row["Symbol"],
                    "asset_id_type": CompanyIDType.SYMBOL,
                }
                for row in rows
            ]
        )

        test_issue_loader = await UpserterT(sa_obj=ai.TestIssue)
        await test_issue_loader.upsert(
            [
                {"asset_id": row["Symbol"], "asset_id_type": CompanyIDType.SYMBOL}
                for row in rows
                if row["Test Issue"] == "Y"
            ]
        )

        status_loader = await UpserterT(sa_obj=ai.Status)
        await status_loader.upsert(
            [
                {
                    "status": financial_status.get(
                        row["Financial Status"], row["Financial Status"]
                    ),
                    "asset_id": row["Symbol"],
                    "asset_id_type": CompanyIDType.SYMBOL,
                }
                for row in rows
            ]
        )


if __name__ == "__main__":
    import asyncio

    from dotenv import load_dotenv

    load_dotenv("/home/dan/danklabs/.localenv")

    asyncio.run(load_nasdaqlisted())
