from functools import partial

from database_utils.upsert_sources import SourcesUpserter
from fin_data import asset_info as ai
from fin_data.asset_info import CompanyIDType
from fin_data.sources.nasdaq.utils import downloaded_file, load_rows
from ready_logger import get_logger

exchange_names = {
    "A": "NYSE MKT",
    "N": "New York Stock Exchange (NYSE)",
    "P": "NYSE ARCA",
    "Z": "BATS Global Markets (BATS)",
    "V": "Investors' Exchange, LLC (IEXG)",
}

logger = get_logger(name="nasdaq-otherlisted")

UpserterT = partial(
    SourcesUpserter.create,
    source="nasdaq otherlisted.txt",
    logger=logger,
)


async def load_otherlisted() -> None:
    with downloaded_file("otherlisted.txt") as file:
        rows, timestamp = load_rows(file)
        logger.info(f"Loaded {len(rows)} rows from file. File update time {timestamp}")

        test_issue_loader = await UpserterT(sa_obj=ai.TestIssue)
        name_loader = await UpserterT(sa_obj=ai.Name)
        exchange_loader = await UpserterT(sa_obj=ai.Exchange)
        type_loader = await UpserterT(sa_obj=ai.Type)

        for symbol_type in ("ACT Symbol", "CQS Symbol", "NASDAQ Symbol"):
            await name_loader.upsert(
                [
                    {
                        "name": row["Security Name"],
                        "asset_id": row[symbol_type],
                        "asset_id_type": CompanyIDType.SYMBOL,
                    }
                    for row in rows
                ]
            )
            await exchange_loader.upsert(
                [
                    {
                        "exchange": exchange_names.get(
                            row["Exchange"], row["Exchange"]
                        ),
                        "asset_id": row[symbol_type],
                        "asset_id_type": CompanyIDType.SYMBOL,
                    }
                    for row in rows
                ]
            )
            await type_loader.upsert(
                [
                    {
                        "type": "ETF",
                        "asset_id": row[symbol_type],
                        "asset_id_type": CompanyIDType.SYMBOL,
                    }
                    for row in rows
                    if row["ETF"] == "Y"
                ]
            )
            await test_issue_loader.upsert(
                [
                    {
                        "asset_id": row[symbol_type],
                        "asset_id_type": CompanyIDType.SYMBOL,
                    }
                    for row in rows
                    if row["Test Issue"] == "Y"
                ]
            )


if __name__ == "__main__":
    import asyncio

    from dotenv import load_dotenv

    load_dotenv("/home/dan/danklabs/.localenv")

    asyncio.run(load_otherlisted())
