import asyncio
from functools import partial

from database_utils.upsert_sources import SourcesUpserter
from fin_data import asset_info as ai
from fin_data.asset_info import CompanyIDType
from fin_data.formatting.exchanges import exchange_name
from fin_data.utils import create_http_client
from ready_logger import get_logger

logger = get_logger(name="nyse")

UpserterT = partial(SourcesUpserter.create, source="nyse.com", logger=logger)


async def load_nyse():
    name_loader = await UpserterT(sa_obj=ai.Name)
    type_loader = await UpserterT(sa_obj=ai.Type)
    exchange_loader = await UpserterT(sa_obj=ai.Exchange)

    async with create_http_client() as client:
        # should have about 750 pages, but checkfor more.
        for page_num in range(1, 1000):
            endpoint = "https://www.nyse.com/api/quotes/filter"
            async with client.post(
                endpoint,
                json={
                    "instrumentType": "EQUITY",
                    "pageNumber": page_num,
                    "sortColumn": "NORMALIZED_TICKER",
                    "sortOrder": "ASC",
                    "maxResultsPerPage": 10,
                    "filterToken": "",
                },
            ) as resp:
                logger.info(f"[{resp.status}] {endpoint}")
                rows = await resp.json()

            if resp and len(rows):
                logger.info(f"Extracted {len(rows)} companies from page {page_num}")

                await name_loader.upsert(
                    [
                        {
                            "name": row["instrumentName"],
                            "asset_id": row["symbolTicker"],
                            "asset_id_type": CompanyIDType.SYMBOL,
                        }
                        for row in rows
                    ]
                )

                await type_loader.upsert(
                    [
                        {
                            "asset_id": row["symbolTicker"],
                            "asset_id_type": CompanyIDType.SYMBOL,
                            "type": row["instrumentType"],
                        }
                        for row in rows
                    ]
                )

                await exchange_loader.upsert(
                    [
                        {
                            "exchange": exchange_name(row["micCode"]),
                            "asset_id": row["symbolTicker"],
                            "asset_id_type": CompanyIDType.SYMBOL,
                        }
                        for row in rows
                    ]
                )

                await asyncio.sleep(1)
            elif page_num > 730:
                logger.info(f"Reached last page: {page_num}")
                break


if __name__ == "__main__":
    import asyncio

    from dotenv import load_dotenv

    load_dotenv("/home/dan/danklabs/.localenv")

    asyncio.run(load_nyse())
