import asyncio
import gzip
import re
from datetime import datetime
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, List

from fin_data.utils import create_http_client
from database_utils import Upserter
from ready_logger import logger
from lxml.html import fromstring
from pytz import timezone
from tqdm import tqdm

from .tables import daily_archive_table, quarterly_archive_table


def remove_header(rows: List[str]) -> List[str]:
    for idx, row in enumerate(rows, start=1):
        if "------------------------------" in row:
            logger.info(f"Removing file header ({idx}/{len(rows)} rows removed)")
            return rows[idx:]
    logger.info("No header found in file.")
    return rows


def parse_file(text: str) -> List[Dict[str, Any]]:
    # split file lines into list and remove file header.
    rows = remove_header([r.strip() for r in text.split("\n")])
    # remove any blank lines.
    rows = [r for r in rows if r != ""]
    # split row text with space separated columns into list.
    rows = [re.sub(r"\s{2,}", "\t", r).split("\t") for r in rows]
    orig_row_count = len(rows)
    # check that each row has 5 columns.
    rows = [r for r in rows if len(r) >= 5]
    if orig_row_count != len(rows):
        logger.warning("%i/%i have column count >= 5", len(rows), orig_row_count)
    db_rows = []
    for row in tqdm(rows):
        *company, filing_type, cik, date, url = row
        db_row = {
            "cik": cik,
            "name": " ".join(company),
            "type": filing_type,
        }
        date_match = re.search(r"(\d{4})-?(\d{2})-?(\d{2})", date)
        if date_match:
            db_row[
                "filing_date"
            ] = f"{date_match.group(1)}-{date_match.group(2)}-{date_match.group(3)}"
        url = url.split("edgar/data/")[-1]
        if url.endswith(".htm"):
            db_row["htm_url"] = url
        elif url.endswith(".txt"):
            db_row["txt_url"] = url
        db_rows.append(db_row)
    return db_rows


async def quarterly_archive(
    start_year: int = 1993, end_year: int = datetime.now().year + 1
) -> None:
    upserter = await Upserter.create(quarterly_archive_table)
    async with create_http_client() as client:
        for year in range(start_year, end_year):
            for quarter in (1, 2, 3, 4):
                quarter_url = f"https://www.sec.gov/Archives/edgar/full-index/{year}/QTR{quarter}/company.idx"
                logger.info("Starting %s quarter %s: %s", year, quarter, quarter_url)
                try:
                    async with client.get(quarter_url) as resp:
                        if resp is None or resp.status == 403:
                            logger.warning(f"Year {year}, quarter {quarter} not found.")
                            continue
                        text = await resp.text()
                        rows = parse_file(text)
                        logger.info("Loading %i rows from %s", len(rows), quarter_url)
                        await upserter.upsert(rows)
                except Exception as e:
                    logger.error(
                        "Error processing quarterly archive page %s: %s", quarter_url, e
                    )


async def extract_gz_text(resp) -> str:
    data = await resp.read()
    file = Path(NamedTemporaryFile().name)
    file.write_bytes(data)
    text = gzip.open(file).read().decode()
    file.unlink()
    return text


async def daily_archive(start_year=1994, end_year=datetime.now().year + 1) -> None:
    upserter = await Upserter.create(daily_archive_table)
    async with create_http_client() as client:
        for year in range(start_year, end_year):
            for quarter in (1, 2, 3, 4):
                quarter_url = f"https://www.sec.gov/Archives/edgar/daily-index/{year}/QTR{quarter}/"
                if (
                    quarter_url
                    == f"https://www.sec.gov/Archives/edgar/daily-index/2014/QTR1/"
                ):
                    continue
                async with client.get(quarter_url) as resp:
                    if resp is None or resp.status == 403:
                        logger.warning(f"Year {year}, quarter {quarter} not found.")
                        continue
                html = await resp.text()
                root = fromstring(html)
                day_urls = [
                    f'{quarter_url}{str(f).strip("/")}'
                    for f in root.xpath('//a[contains(@href,"company.")]/@href')
                ]
                for day_url in day_urls:
                    async with client.get(day_url) as resp:
                        if day_url.endswith(".gz"):
                            text = await extract_gz_text(resp)
                        else:
                            text = await resp.text()
                    rows = parse_file(text)
                    await upserter.upsert(rows)


async def wait_for_todays_archive():
    # get current year
    year = datetime.now().year
    year_url = f"https://www.sec.gov/Archives/edgar/daily-index/{year}/"
    async with create_http_client() as client:
        async with client.get(year_url) as resp:
            text = await resp.text()
        root = fromstring(text)
        quarter_url = year_url + str(
            root.xpath('//a[contains(@href, "QTR")]/@href')[-1]
        )
        today = datetime.now(timezone("America/New_York")).strftime("%Y%m%d")
        url = f"{quarter_url}company.{today}.idx"
        while True:
            async with client.head(year_url) as resp:
                if resp.status == 403:
                    logger.info(f"Waiting for today's daily archive at {url}")
                    await asyncio.sleep(60)
                else:
                    logger.info(f"Found today's daily archive at {url}")
                    break
        async with client.get(url) as resp:
            text = await resp.text()
            rows = parse_file(text)
    upserter = await Upserter.create(daily_archive_table)
    await upserter.upsert(rows)
