import json
import re
from typing import Any, Dict

from ready_logger import logger
from lxml.html import fromstring

from .utils import (
    comma_filter,
    comma_join,
    extract_first,
    no_matching_symbol,
    space_join,
)

info_box_field_regs = {
    "state": r"State location[\s.]*:",
    "state_of_incorp": r"State of (?:Inc|Incorp|Incorporation)[\s.]*:",
    "fiscal_year_end": r"Fiscal Year End[\s.]*:",
    "formerly": r"formerly[\s.]*:",
    "sic": r"SIC[\s.]*:",
    "office": r"\(Office of",
}
info_box_separators = ["\|"]
info_box_separators.extend(info_box_field_regs.values())
info_box_separators = f"({'|'.join(info_box_separators)})"
# value starts at end of key and ends at separator.
info_box_field_regs = {
    field: re.compile(f"(?i){field_reg}(.+?){info_box_separators}")
    for field, field_reg in info_box_field_regs.items()
}

phone_num_re = re.compile(r"(\s?-\s?|\s|\.|\))".join([r"\(?\d{3}", r"\d{3}", r"\d{4}"]))


def get_info_box_text(root):
    """
    Extract and clean company info text.
    Works on both search results page and filing page.
    """
    info_box_text = space_join(root.xpath('//p[@class="identInfo"]//text()'))
    # remove unwanted text.
    return re.sub(
        r"(?i)Get insider transactions for this (issuer|reporting owner)\s?\.?",
        "",
        info_box_text,
    )


def parse_company_info_text_box(root):
    """
    Extract company info from company info text.
    Works on both search results page and filing page.
    """
    info_box_text = get_info_box_text(root)

    comp_info = {}
    for field, field_reg in info_box_field_regs.items():
        match = [m[0].strip("), ") for m in field_reg.findall(info_box_text)]
        match = [m for m in match if m != ""]
        if len(match) > 1:
            logger.warning(
                f"Multiple ({len(match)}) matches for company info box key '{field}'"
            )
            comp_info[field] = json.dumps(match)
        elif len(match) == 1:
            comp_info[field] = match.pop()
    return comp_info


async def get_company_info(symbol: str, session) -> Dict[str, Any]:
    """
    Extract data from company info box at top of page.
    Works on both search results page and filing page.
    """
    resp = await session.get(
        f"https://www.sec.gov/cgi-bin/browse-edgar?action=getcompany&CIK={symbol}"
    )
    html = await resp.text()
    if await no_matching_symbol(html, symbol):
        return
    root = fromstring(html)
    comp_info = {
        "symbol": symbol,
        **parse_company_info_text_box(root),
    }
    # extract remaining company info from HTML.
    comp_name = root.xpath('//span[@class="companyName"]/text()')
    if len(comp_name):
        comp_info["name"] = space_join(comp_name).strip().strip("# ")

    cik = root.xpath(
        '//acronym[@title="Central Index Key"]/following-sibling::a/text()'
    )
    if len(cik):
        cik = extract_first(cik)
        comp_info["cik"] = (
            re.sub(r"(?i)\(see all company filings\)", "", cik).strip().lstrip("0")
        )

    mailing_address = root.xpath(
        '//div[@class="mailer"][contains(text(),"Mailing Address")]/span[@class="mailerAddress"]/text()'
    )
    if len(mailing_address):
        comp_info["mailing_address"] = comma_join(mailing_address)

    bussiness_addr_and_phone = root.xpath(
        '//div[@class="mailer"][contains(text(),"Business Address")]/span[@class="mailerAddress"]/text()'
    )
    if len(bussiness_addr_and_phone):
        bussiness_addr_and_phone = comma_join(bussiness_addr_and_phone)
        phone_num = set(
            m.group() for m in phone_num_re.finditer(bussiness_addr_and_phone)
        )
        if len(phone_num):
            comp_info["phone"] = ", ".join(phone_num)
        comp_info["bussiness_address"] = comma_filter(
            phone_num_re.sub("", bussiness_addr_and_phone)
        )
    return comp_info
