import re

from database_utils.upsert_sources import SourcesUpserter
from fin_data import asset_info as ai
from fin_data.asset_info import CompanyIDType
from fin_data.utils import create_http_client, user_header
from ready_logger import get_logger


async def sec_cik_lookup_list():
    """(13 MB, text file) Note that this list includes funds and individuals and is historically cumulative for company names.
    Thus a given CIK may be associated with multiple names in the case of company or fund name changes,
    and the list contains some entities that no longer file with the SEC.
    """
    logger = get_logger(name="sec-cik-lookup-list")
    url = "https://www.sec.gov/Archives/edgar/cik-lookup-data.txt"
    logger.info(f"Fetching file: {url}")
    async with create_http_client() as client:
        async with client.get(url, headers=user_header()) as resp:
            logger.info(f"[{resp.status}] {url}")
            if resp.status != 200:
                raise RuntimeError(f"Bad Response ({resp.status}): {url}")
            text = await resp.text()

    logger.info("Parsing file.")
    rows = [row.strip(" :").split(":") for row in text.split("\n")]
    count_unchecked = len(rows)
    rows = [row for row in rows if len(row) > 1]
    if missing_colon := (len(rows) - count_unchecked):
        logger.error(
            f"{missing_colon}/{count_unchecked} rows are missing colon delimiter and can not be parsed."
        )
    rows = [
        {
            "cik": row.pop().lstrip("0"),
            "name": re.sub(r"\s+", " ", " ".join(row)).strip(),
        }
        for row in rows
    ]
    logger.info(f"Loaded {len(rows)} rows from file.")

    cik_loader = await SourcesUpserter.create(
        sa_obj=ai.CIK,
        source="sec-cik-lookup-list",
        logger=logger,
    )
    await cik_loader.upsert(
        [
            {
                "cik": row["cik"],
                "asset_id": row["name"],
                "asset_id_type": CompanyIDType.NAME,
            }
            for row in rows
        ]
    )

    name_loader = await SourcesUpserter.create(
        sa_obj=ai.Name,
        source="sec-cik-lookup-list",
        logger=logger,
    )
    await name_loader.upsert(
        [
            {
                "name": row["name"],
                "asset_id": row["cik"],
                "asset_id_type": CompanyIDType.CIK,
            }
            for row in rows
        ]
    )
    logger.info(f"Finished loading {len(rows)} rows to database.")


if __name__ == "__main__":
    import asyncio

    from dotenv import load_dotenv

    load_dotenv("/home/dan/danklabs/.localenv")

    asyncio.run(sec_cik_lookup_list())
