import asyncio
import re
from functools import partial

from database_utils.upsert_sources import SourcesUpserter
from fin_data import asset_info as ai
from fin_data.asset_info import CompanyIDType
from fin_data.utils import create_http_client, user_header
from ready_logger import get_logger


async def edgar_search_list():
    """
    json file for ticker/CIK/Company name associations.
    These associations provide the typeahead lookup and search functionality for the home page EDGAR search.
    The company names are EDGAR conformed names. We periodically update the file but do not guarantee accuracy or scope.
    """
    logger = get_logger(name="sec_edgar_search_list")

    UpserterT = partial(
        SourcesUpserter.create,
        source="sec edgar search list",
        logger=logger,
    )
    name_loader = await UpserterT(sa_obj=ai.Name)
    cik_loader = await UpserterT(sa_obj=ai.CIK)

    url = "https://www.sec.gov/files/company_tickers.json"
    async with create_http_client() as client:
        async with client.get(url, headers=user_header()) as resp:
            logger.info(f"[{resp.status}] {url}")
            data = await resp.json()

    rows = [row for _, row in data.items()]
    symbols = [row["ticker"].strip().upper() for row in rows]

    ciks = [str(row["cik_str"]).lstrip("0") for row in rows]
    await cik_loader.upsert(
        [
            {
                "cik": cik,
                "asset_id": symbol,
                "asset_id_type": CompanyIDType.SYMBOL,
            }
            for symbol, cik in zip(symbols, ciks)
        ]
    )

    names = [re.sub(r"\s+", " ", row["title"].strip()) for row in rows]
    await name_loader.upsert(
        [
            {
                "name": name,
                "asset_id": symbol,
                "asset_id_type": CompanyIDType.SYMBOL,
            }
            for symbol, name in zip(symbols, names)
        ]
    )


if __name__ == "__main__":
    import asyncio

    from dotenv import load_dotenv

    load_dotenv("/home/dan/danklabs/.localenv")

    asyncio.run(edgar_search_list())
