import asyncio

from database_utils.upsert_sources import SourcesUpserter
from fin_data import asset_info as ai
from fin_data.asset_info import CompanyIDType
from fin_data.utils import create_http_client, user_header
from ready_logger import get_logger


async def load_ticker_txt_list() -> None:
    """
    tab-delimited text file (ticker.txt) for ticker / CIK.
    This file provides the ticker search functionality for the Company Search Fast Search.
    We periodically update the file but do not guarantee accuracy or scope.
    """
    logger = get_logger(name="sec_ticket_txt_list")

    url = "https://www.sec.gov/include/ticker.txt"
    async with create_http_client() as client:
        async with client.get(url, headers=user_header()) as resp:
            logger.info(f"[{resp.status}] {url}")
            text = await resp.text()

    rows = [r.strip() for r in text.split("\n")]
    rows = [r for r in rows if r != ""]
    symbol_cik = []
    for row in rows:
        row = row.split("\t")
        if len(row) == 2:
            symbol_cik.append(row)
        else:
            logger.error(f"Could not parse row {row}. Expencted row of length 2.")

    cik_loader = await SourcesUpserter.create(
        sa_obj=ai.CIK,
        source="sec-ticker-txt-list",
        logger=logger,
    )
    await cik_loader.upsert(
        [
            {
                "cik": cik.strip().lstrip("0"),
                "asset_id": symbol.strip().upper(),
                "asset_id_type": CompanyIDType.SYMBOL,
            }
            for symbol, cik in symbol_cik
        ]
    )


if __name__ == "__main__":
    import asyncio

    from dotenv import load_dotenv

    load_dotenv("/home/dan/danklabs/.localenv")

    asyncio.run(load_ticker_txt_list())
