import asyncio
import re
from datetime import datetime
from typing import Any, Dict

import aioredis
import sqlalchemy as sa
from fin_data.utils import create_http_client
from database_utils import Upserter, data_id_creator
from ready_logger import logger
from lxml.html import fromstring
from tqdm import tqdm

from .tables import (
    PageType,
    failed_page_loads_table,
    filings_table,
    last_filing_searches,
)
from .utils import comma_join, extract_first, no_matching_cik, rate_limiter, space_join


class FilingsMetadataSearch:
    @classmethod
    async def create(cls, redis_url: str):
        self = cls()
        self.filings_search_columns = {c.name for c in filings_table.columns.values()}
        self.filings_table_upserter = await Upserter.create(sa_obj=filings_table)
        self.failed_page_loads_table_upserter = await Upserter.create(
            sa_obj=failed_page_loads_table
        )
        self.last_filing_searches_upserter = await Upserter.create(
            sa_obj=last_filing_searches,
            column_values={
                "times_failed": sa.text(
                    f'{failed_page_loads_table.schema}."{failed_page_loads_table.name}".times_failed + 1'
                )
            },
        )
        self._create_filings_row_id = data_id_creator(filings_table)
        self._redis = aioredis.from_url(redis_url)
        self._http_client = create_http_client()
        return self

    async def search_all_queued_ciks(self, filing_type: str, batch_size: int) -> None:
        """Search {filing_type} for all queued CIKs

        Args:
            filing_type (str): The type of filing to search for (e.g. 10-K)
            batch_size (int): Number of CIKs to search simultaneously.
        """
        logger.info(f"Searching filing type: {filing_type}")

        while True:
            cik_batch = await self._redis.spop(
                f"{filing_type}_ciks_not_searched", count=batch_size, encoding="utf-8"
            )
            if not cik_batch:
                logger.info("All CIKs have been processed.")
                break
            await asyncio.gather(
                *[
                    asyncio.create_task(self.search_cik_type(cik, filing_type))
                    for cik in cik_batch
                ]
            )
            logger.info(f"Finished CIK batch of {batch_size}")
        logger.info(f"Finished searching all CIKs for filing type {filing_type}")

    async def _try_get(self, url: str, cik: str, filing_type: str, page_type: PageType):
        await rate_limiter.maybe_sleep(url)
        try:
            return await self.http_client.get(url)
        except Exception as e:
            logger.error(f"Error fetching page {url}: {e}")
            await self.failed_page_loads_table_upserter.upsert(
                {
                    "url": url,
                    "cik": str(cik).lstrip("0"),
                    "filing_type": str(filing_type).lower(),
                    "page_type": page_type,
                    "last_fail": datetime.utcnow(),
                }
            )

    def get_filing_metadata(self, root):
        """Extract filing metadata from the top of a filing page."""
        filing_metadata = {}
        filing_date = root.xpath(
            '//div[contains(text(),"Filing Date")]/following-sibling::div/text()'
        )
        if len(filing_date):
            filing_metadata["filing_date"] = extract_first(filing_date)
        accepted_date = root.xpath(
            '//div[contains(text(),"Accepted")]/following-sibling::div/text()'
        )
        if len(accepted_date):
            filing_metadata["accepted_date"] = extract_first(accepted_date)
        doc_count = root.xpath(
            '//div[contains(text(),"Documents")]/following-sibling::div/text()'
        )
        if len(doc_count):
            filing_metadata["doc_count"] = extract_first(doc_count)
        period_of_report = root.xpath(
            '//div[contains(text(),"Period of Report")]/following-sibling::div/text()'
        )
        if len(period_of_report):
            filing_metadata["period_of_report"] = extract_first(period_of_report)
        sec_accession_num = root.xpath('//div[@id="secNum"]/text()')
        if len(sec_accession_num):
            filing_metadata["sec_accession_num"] = space_join(sec_accession_num)
        return filing_metadata

    def extract_documents_tables(self, root):
        """Extract (from filing page) metadata of all documents that compose a filing."""

        def is_header_row(row):
            "Return True if row is header of documens table."
            header_text = space_join(row.xpath(".//text()"))
            # standard metadata fields used on company filing page tables.
            present_fields = [
                t
                for t in ("seq", "description", "document", "type", "size")
                if t in header_text.lower()
            ]
            return len(present_fields) > 2

        # should have two tables: 'Document Format Files' and 'Data Files'. Document Format Files has the filings.
        tables = root.xpath('//table[@class="tableFile"]')
        if not len(tables):
            return []
        table = tables[0]
        rows = table.xpath(".//tr")
        # rows should be header followed by rows of filings metadata.
        if len(rows) and is_header_row(rows[0]):
            # remove header.
            del rows[0]
        documents = []
        for idx, row in enumerate(rows):
            columns = row.xpath("./td")
            if len(columns) != 5:
                logger.error(
                    f"Company filing row {idx} has {len(columns)} columns. Expected 5. Row: {space_join(row.xpath('.//text()'))}"
                )
                continue
            document = {}
            url = row.xpath('.//a[contains(@href,"/Archives/edgar/data")]/@href')
            if len(url):
                document["url"] = extract_first(url)
            seq = columns[0].xpath("./text()")
            if len(seq):
                document["seq"] = extract_first(seq)
            description = columns[1].xpath("./text()")
            if len(description):
                document["description"] = extract_first(description)
            doc_type = columns[3].xpath("./text()")
            if doc_type:
                document["type"] = extract_first(doc_type)
            size = columns[4].xpath("./text()")
            if len(size):
                try:
                    document["size"] = int(extract_first(size))
                except ValueError:
                    pass
            documents.append(document)
        return documents

    def get_filing_txt_url(self, documents, filing_page_url):
        """Extract final submission .txt URL from filing page."""
        # score each .txt doc
        doc_sizes = [d["size"] for d in documents if "size" in d]
        doc_sizes.sort()
        txt_scores = []
        for idx, doc in enumerate(documents):
            url = doc.get("url", "")
            if url.endswith(".txt"):
                score = 0
                # 'complete submission text' should be in description.
                if "complete submission text" in doc.get("description", "").lower():
                    score += 3
                # Seq should be empty.
                if not doc.get("seq"):
                    score += 1
                # Type should be empty.
                if not doc.get("type"):
                    score += 1
                # complete txt file should be last in list.
                if idx == len(documents) - 1:
                    score += 1
                # stem of file name url should be in page url.
                if url.split("/")[-1].rstrip(".txt") in filing_page_url:
                    score += 1
                # Size should be one of the largest.
                # larger size should yield larger ratio.
                if (
                    "size" in doc
                    and (doc_sizes.index(doc["size"]) + 1) / len(doc_sizes) > 0.65
                ):
                    score += 1
                txt_scores.append((score, url))
        if len(txt_scores):
            txt_scores.sort(key=lambda x: x[0])
            score, url = txt_scores.pop()
            # return score fraction of max possible points.
            return round(score / 8, 2), url

    def get_filing_htm_url(self, documents, filing_type):
        """Extract final submission .htm URL from filing page."""

        def filter_filing_type(filing_type):
            return filing_type.replace("-", "").replace("/", "").lower()

        # score each .htm file
        filing_type = filter_filing_type(filing_type)
        doc_sizes = [d["size"] for d in documents if "size" in d]
        doc_sizes.sort()
        htm_scores = []
        for idx, doc in enumerate(documents):
            url = doc.get("url", "")
            if re.search(r"\.(html?|xml)$", url):
                score = 0
                # index should be as beginning.
                if idx < 2:
                    score += 1
                if url.startswith("/ix?doc="):
                    score += 1
                # field 'type' should start with filing_type if file is htm.
                if filing_type in filter_filing_type(doc.get("type", "")):
                    score += 4
                # field 'description' should start with filing_type if file is htm.
                if (
                    filing_type in filter_filing_type(doc.get("description", ""))
                    or "PRIMARY DOCUMENT" in doc.get("description", "").upper()
                ):
                    score += 2
                # File name should contain filing type if file is htm.
                if filing_type in filter_filing_type(url):
                    score += 1
                # Seq 1 is usually complete .htm submission.
                if doc.get("seq") == "1":
                    score += 1
                # Size should be one of the largest or left blank.
                # larger size should yield larger ratio.
                if (
                    "size" in doc
                    and (doc_sizes.index(doc["size"]) + 1) / len(doc_sizes) > 0.65
                    or str(doc.get("size", "")).strip() == ""
                ):
                    score += 1
                htm_scores.append((score, url))
        if len(htm_scores):
            htm_scores.sort(key=lambda x: x[0])
            score, url = htm_scores.pop()
            # return score fraction of max possible points.
            return round(score / 11, 2), url

    def get_filing_page_data(self, root, url, filing_type):
        "Extract all data from a filing page."
        data = self.get_filing_metadata(root)
        documents = self.extract_documents_tables(root)
        if documents:
            data["documents"] = documents
            filing_htm = self.get_filing_htm_url(documents, filing_type)
            if filing_htm:
                htm_score, htm_url = filing_htm
                data["htm_url"] = htm_url
                data["htm_url_score"] = htm_score
            filing_txt = self.get_filing_txt_url(documents, url)
            if filing_txt:
                txt_score, txt_url = filing_txt
                data["txt_url"] = txt_url
                data["txt_url_score"] = txt_score
        return data

    def parse_search_results_table(self, root):
        """Extract filing metadata for each filing in search results table."""

        def is_header_row(row_ele):
            """Return True if row_ele is table header of a company filing page table."""
            # standard metadata fields used on company filing page tables.
            row_text = space_join(row_ele.xpath(".//text()"))
            present_fields = [
                t
                for t in (
                    "filings",
                    "format",
                    "description",
                    "filing date",
                    "file/film number",
                )
                if t in row_text.lower()
            ]
            return len(present_fields) > 2

        # list of root elements for table rows.
        rows = root.xpath('//table[@summary="Results"]//tr')
        # rows should be header followed by rows of filings metadata.
        if len(rows) and is_header_row(rows[0]):
            # remove header.
            del rows[0]
        rows_data = []
        for row in tqdm(rows):
            row_cols = row.xpath("./td")
            if len(row_cols) != 5:
                logger.error(
                    f"Company filing row with {len(row_cols)} columns (Expected 5)."
                )
                continue
            filing_meta = {}
            filing_type = row_cols[0].xpath("./text()")
            if len(filing_type):
                filing_meta["type"] = extract_first(filing_type)
            description = row_cols[2].xpath(".//text()")
            if len(description):
                filing_meta["description"] = comma_join(description)
            filing_date = row_cols[3].xpath("./text()")
            if len(filing_date):
                filing_meta["filing_date"] = extract_first(filing_date)
            file_film_num = row_cols[4].xpath(".//text()")
            if len(file_film_num):
                filing_meta["file_film_num"] = comma_join(file_film_num)
            documents_url = row.xpath('.//a[@id="documentsbutton"]/@href')
            if len(documents_url):
                filing_meta["documents_url"] = extract_first(documents_url)
            interactive_url = row.xpath('.//a[@id="interactiveDataBtn"]/@href')
            if len(interactive_url):
                filing_meta["interactive_url"] = extract_first(interactive_url)
            rows_data.append(filing_meta)
        logger.info(f"Extracted {len(rows_data)} filing metadata rows.")
        return rows_data

    async def parse_filing_page(self, filing_row: Dict[str, Any]) -> None:
        # navigate to filing page and extract data.
        url = filing_row["documents_url"]
        if not re.match(r"https?://www.sec.gov", url):
            url = url.lstrip("/")
            if not url.startswith("edgar/data/"):
                url = "edgar/data/" + url
            url = "http://www.sec.gov/" + url
        if (
            resp := self._try_get(
                url, filing_row["cik"], filing_row["type"], PageType.FILING
            )
        ) is not None:
            html = await resp.text()
            filing_page_data = self.get_filing_page_data(
                fromstring(html), filing_row["documents_url"], filing_row["type"]
            )
            # update filing meta with data from filing page.
            filing_row.update(filing_page_data)
            filing_row = {
                k: v for k, v in filing_row.items() if k in filings_search_columns
            }
            if "cik" in filing_row:
                filing_row["cik"] = filing_row["cik"].lstrip("0")
            if "htm_url" in filing_row:
                filing_row["htm_url"] = filing_row["htm_url"].split("edgar/data/")[-1]
            if "txt_url" in filing_row:
                filing_row["txt_url"] = filing_row["txt_url"].split("edgar/data/")[-1]
            filing_row["documents_url"] = filing_row["documents_url"].split(
                "edgar/data/"
            )[-1]
            filing_row["data_id"] = self._create_filings_row_id(filing_row)
            await self.filings_table_upserter.upsert([filing_row])

    async def search_cik_type(self, cik: str, filing_type: str):
        logger.info(f"Searching CIK '{cik}'. Filing type '{filing_type}'")
        has_filing_type = str(filing_type) not in ("none", "null", "")
        # should not have more than a couple pages. if high number of pages, page probably has button path w/o real button.
        page_num = 0
        while True:
            url = f"https://www.sec.gov/cgi-bin/browse-edgar?action=getcompany&CIK={cik}&start={100*page_num}&count=100"
            if has_filing_type:
                url += f"&type={filing_type}"
            logger.info(f"Starting {cik} {filing_type} page {page_num}: {url}")

            # load page.
            resp = await self.http_client.get(url)
            if resp is None:
                await self.record_failed_page_load(
                    url, cik, filing_type, PageType.SEARCH_RESULTS
                )
            html = await resp.text()
            if await no_matching_cik(html, cik):
                return
            root = fromstring(html)

            filing_rows = self.parse_search_results_table(root)
            for filing_row in filing_rows:
                filing_row["cik"] = cik
                filing_row["type"] = filing_row.get("type") or filing_type
                await self.parse_filing_page(filing_row)

            # return if there are no more pages.
            if not len(root.xpath('//input[@type="button"][contains(@value,"Next")]')):
                logger.info(
                    f"Processed all {page_num+1} search result pages for CIK {cik}"
                )
                # mark time filing search was finished.
                await self.last_filing_searches_upserter.upsert(
                    {
                        "cik": cik,
                        "filing_type": str(filing_type).lower(),
                        "last_search": datetime.utcnow(),
                    }
                )
                logger.info(f"Finished search {filing_type} for CIK {cik}")
                return
            page_num += 1
