from utils import get_logger, get_mysql_conn, get_s3_conn, get_redis_conn

from distbot.spider import Spider
from PyPDF2 import PdfFileReader, PdfFileWriter
import aioredis
import requests
import boto3

from typing import Union, List, Dict, Any, Optional
from datetime import datetime
from statistics import mean
from pathlib import Path
from time import time
import argparse
import asyncio
import signal
import json
import re

logger = get_logger()

def read_pdf_metadata(file: Path):
    with file.open(mode='rb') as i:
        reader = PdfFileReader(i)
        metadata = reader.getDocumentInfo()
        data = metadata.get('/Subject')
        if data:
            return json.loads(data)


def write_pdf_metadata(file, filing_meta):
    # convert datetime to str
    if 'created' in filing_meta:
        filing_meta['created'] = str(filing_meta['created'])
    # store metadata as JSON
    filing_meta = json.dumps(filing_meta)
    writer = PdfFileWriter()
    with file.open(mode='rb') as i:
        try:
            reader = PdfFileReader(i)
            writer.appendPagesFromReader(reader)
            metadata = reader.getDocumentInfo()
            writer.addMetadata(metadata)
            writer.addMetadata({"/Subject": filing_meta})
            with file.open(mode='ab') as o:
                writer.write(o)
        except Exception as e:
            logger.error(f"Error writing metadata to PDF {file}: {e}")


class PDFGenerator:
    def __init__(self, filing_meta_key: str,
                 existing_files_key: str,
                 wait_for_queued: bool = False,
                 save_dir: Optional[Path] = None,
                 browser: Optional[str] = None):
        self.filing_meta_key = filing_meta_key
        self.existing_files_key = existing_files_key
        self.wait_for_queued = wait_for_queued
        self.save_dir = Path(
            save_dir) if save_dir is not None else Path('pdf_tmp')
        self.save_dir.mkdir(exist_ok=True, parents=True)
        self.launch_options = {
            "headless": True,
            "ignoreHTTPSErrors": True,
            "defaultViewport": {},
            "defaultNavigationTimeout": 60_000*3,
            "pageIdleTimeout": 60*4,
            "args": [
                "--disable-web-security",
                "--no-sandbox",
                "--start-maximized",
                "--disable-gpu",
            ]
        }
        if browser:
            self.launch_options['executablePath'] = browser
        self.spider = None
        self.s3_conn = get_s3_conn()
        self.mysql_conn = get_mysql_conn()
        self.loop = asyncio.get_event_loop()
        self.loop.run_until_complete(asyncio.gather(
            self.set_spider(),
            self.set_redis_conn__()))
        logger.info("Initialized connections.")
        for s in (signal.SIGINT, signal.SIGTERM):
            self.loop.add_signal_handler(
                s, lambda s=s: self.loop.create_task(self.shutdown(s)))
        self.loop.create_task(self.process_queue())
        self.loop.run_forever()

    async def set_redis_conn__(self):
        self.redis_conn = await get_redis_conn()
        # TODO change this
        self.redis_conn2 = await aioredis.create_redis("redis://localhost:6379/0")

    async def shutdown(self, signame=None):
        logger.info(f"Caught signal {signame}. Shutting down.")
        self.redis_conn.close()
        self.mysql_conn.close()
        await self.spider.shutdown()
        tasks = [t for t in asyncio.all_tasks(
        ) if t is not asyncio.current_task()]
        [task.cancel() for task in tasks]
        await asyncio.gather(*tasks, return_exceptions=True)
        self.loop.stop()
        logger.info("Shutdown complete.")

    async def set_spider(self):
        logger.info("Setting spider.")
        if self.spider is not None:
            await self.spider.shutdown()
        self.spider = Spider()
        await self.spider.add_browser(launch_options=self.launch_options)

    async def create_pdf(self, url, filing_meta, save_path, retries=1) -> bool:
        async def try_create_pdf(url: str, save_path: Union[str, Path]):
            # navigate to filing page.
            resp, page = await self.spider.get(url)
            start = time()
            if resp.status != 200:
                if resp.status == 429:
                    await asyncio.sleep(60*11)
                raise Exception(f"Bad response: {resp.status}")
            # generate pdf.
            await page.pdf(path=str(save_path))
            logger.info(f"Created PDF ({time()-start} s): {str(save_path)}")
            await self.spider.set_idle(page)
            # add metadata to pdf.
            write_pdf_metadata(save_path, filing_meta)
        for _ in range(retries):
            try:
                await asyncio.wait_for(
                    try_create_pdf(url, save_path), timeout=60*2)
                return True
            except asyncio.TimeoutError:
                logger.error(f"TimeoutError creating PDF: {filing_meta}")
                await self.set_spider()
            except Exception as e:
                logger.error(f"Error creating PDF: {e}")
        logger.error(
            f"Exceeded max retries ({retries}) when generating PDF of {url}")
        return False

    async def get_file_name(self, cik: str, filing_type: str, filing_date: str):
        cik = cik.lstrip('0')
        filing_type = filing_type.replace('/', '-')
        filing_date = re.search(r'\d{4}-\d{2}-\d{2}', filing_date).group()
        file_no = 1
        while True:
            stem = f"{cik}_{filing_type}_{filing_date}_F{file_no}"
            exists = await self.redis_conn2.sismember(self.existing_files_key, stem)
            if not exists and not Path(f'/media/dan/easystore/{stem}.pdf').is_file() and not Path(f'/media/dan/easystore{stem}.htm').is_file():
                await self.redis_conn2.sadd(self.existing_files_key, stem)
                return stem+'.pdf'
            file_no += 1

    async def download_filing(self, filing_meta: Dict[str, Any]):
        url = filing_meta['htm_url'] if 'htm_url' in filing_meta and float(
            filing_meta['htm_url_score']) > 0.59 else filing_meta['txt_url']
        url = f"https://www.sec.gov/Archives/edgar/data/{url}"
        file_name = await self.get_file_name(filing_meta['cik'], filing_meta['type'], filing_meta['filing_date'])
        save_path = self.save_dir.joinpath(file_name)
        created_ok = await self.create_pdf(url, filing_meta, save_path)
        if not created_ok:
            # save HTML file instead.
            # TODO aiohttp?
            resp = requests.get(url)
            if resp.status_code == 200:
                # remove .pdf suffix
                save_path = save_path.with_suffix('.htm')
                logger.info(f"Saving HTML for {url} -> {save_path}")
                save_path.write_text(resp.text)
            else:
                logger.error(
                    f"Non-200 response code ({resp.status_code}) when downloading HTML of {url}")
                return
        # update file in database.
        self.mysql_conn.query(
            f"UPDATE filings_search SET file='{save_path.name}' WHERE cik='{filing_meta['cik']}' AND type='{filing_meta['type']}' AND filing_date='{filing_meta['filing_date']}' AND txt_url='{filing_meta['txt_url']}'")
        self.mysql_conn.commit()
        # upload to s3.
        self.s3_conn.Bucket('sec-pdfs').put_object(Key=save_path.name,
                                                   Body=save_path.read_bytes())
        if self.save_dir.name == 'pdf_tmp':
            # remove file from disk
            save_path.unlink()

    async def process_queue(self):
        filing_meta = await self.redis_conn.spop(self.filing_meta_key, encoding='utf-8')
        if filing_meta is None:
            if not self.wait_for_queued:
                logger.info("ALL FILINGS ARE DOWNLOADED.")
                await self.shutdown()
                return
            logger.info("No queued filings. Waiting.")
            await asyncio.sleep(1)
        filing_meta = json.loads(filing_meta)
        await self.download_filing(filing_meta)
        self.loop.create_task(self.process_queue())


def mean_pdf_creation_rate(log_text: str):
    times = [datetime.fromisoformat(m.group(1)) for m in re.search(
        r'(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}),\d+: Created PDF', log_text, re.MULTILINE)]
    return mean(times[next_idx] - time for next_idx, time in enumerate(times[:-1], start=1))


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--filing_meta_key', '-k', type=str, required=True,
                        help='Key to Redis Set where filings metadata JSON is cached.')
    parser.add_argument('--existing_files_key', '-f', type=str, required=True,
                        help='Key to Redis Set containing names of all existing filing PDFs.')
    parser.add_argument('--wait_for_queued', '-w', action='store_true',
                        help="Poll filing_meta_key when empty, rather than exiting.")
    parser.add_argument('--save_dir', '-d', type=str,
                        help='Local disk directory where PDFs will be saved.')
    parser.add_argument('--browser', '-b', type=str, default='/bin/google-chrome-stable',
                        help='Path to browser executable (Chrome or Chromium).')
    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()
    PDFGenerator(filing_meta_key=args.filing_meta_key,
                 existing_files_key=args.existing_files_key,
                 wait_for_queued=args.wait_for_queued,
                 save_dir=args.save_dir,
                 browser=args.browser)
