from datetime import datetime

import feedparser
from aiohttp import ClientSession
from fin_data.utils import create_http_client
from feedparser.util import FeedParserDict
from ready_logger import get_logger
from tqdm import tqdm

from ..filings_search import FilingsMetadataSearch
from ..utils import rate_limiter
from .utils import parse_standard_feed

logger = get_logger("Historic RSS")


async def _get_feed(url: str, client: ClientSession) -> FeedParserDict:
    await rate_limiter.maybe_sleep(url)
    try:
        async with client.get(url) as resp:
            html = await resp.text()
        return feedparser.parse(html)
    except Exception as e:
        logger.error("Could not parse feed at %s: %s", url, e)


async def parse_mutual_funds():
    # Mutual fund filings containing a risk/return summary tagged using the US Mutual Fund Risk/Return Taxonomy
    # This is a list of up to 200 of the latest filings containing a risk/return summary tagged using the US Mutual Fund Risk/Return Taxonomy, updated every 10 minutes.
    # resp = session.get('https://www.sec.gov/Archives/edgar/xbrl-rr.rss.xml')
    pass


async def all_monthly_rss(redis_url: str):
    logger.info("Parsing monthly RSS")
    urls = [
        "https://www.sec.gov/Archives/edgar/monthly/xbrlrss-2005-04.xml",
        "https://www.sec.gov/Archives/edgar/monthly/xbrlrss-2005-05.xml",
        "https://www.sec.gov/Archives/edgar/monthly/xbrlrss-2005-07.xml",
        "https://www.sec.gov/Archives/edgar/monthly/xbrlrss-2005-08.xml",
        "https://www.sec.gov/Archives/edgar/monthly/xbrlrss-2005-09.xml",
        "https://www.sec.gov/Archives/edgar/monthly/xbrlrss-2005-10.xml",
        "https://www.sec.gov/Archives/edgar/monthly/xbrlrss-2005-11.xml",
        "https://www.sec.gov/Archives/edgar/monthly/xbrlrss-2005-12.xml",
    ]
    current_year = datetime.now().year
    urls += [
        f"https://www.sec.gov/Archives/edgar/monthly/xbrlrss-{year}-{month:02}.xml"
        for year in range(2006, current_year)
        for month in range(1, 13)
    ]
    urls += [
        f"https://www.sec.gov/Archives/edgar/monthly/xbrlrss-{current_year}-{month:02}.xml"
        for month in range(1, datetime.now().month)
    ]
    meta_searcher = await FilingsMetadataSearch.create(redis_url=redis_url)
    async with create_http_client() as client:
        for url in tqdm(urls):
            if feed := _get_feed(url, client):
                entries = await parse_standard_feed(feed)
                for row in tqdm(entries):
                    await meta_searcher.parse_filing_page(row)
