import asyncio
import re
from datetime import datetime
from typing import Any, Dict, List

import aioredis
import feedparser
import sqlalchemy as sa
from fin_data.sources.sec.filings_search import FilingsMetadataSearch
from fin_data.sources.sec.tables import daily_rss_table, rss_last_update_table
from fin_data.sources.sec.utils import rate_limiter
from fin_data.utils import create_http_client
from database_utils import Upserter, get_engine
from feedparser.util import FeedParserDict
from ready_logger import get_logger
from pytz import timezone
from tqdm import tqdm

from .utils import parse_standard_feed


class RSSStreamParser:
    logger = get_logger("RSSStreamParser")

    async def create(cls, redis_url: str) -> "RSSStreamParser":
        self = cls()
        self._http_client = create_http_client()
        self._redis = aioredis.from_url(redis_url)
        self._engine = get_engine()
        self._daily_rss_upserter = await Upserter.create(sa_obj=daily_rss_table)
        self._meta_searcher = await FilingsMetadataSearch.create(redis_url=redis_url)
        return self

    async def run(self) -> None:
        self.logger.info("Starting.")
        while True:
            await self.parse_feeds_and_filing_pages()
            await asyncio.sleep(1)

    async def parse_feeds_and_filing_pages(self) -> None:
        """for each filing (as specified by documents_url), combine column data from both 10 min and latest (should be the same)"""
        latest_feed_entries = await self.parse_latest_feed()
        ten_minute_rss_entries = await self.parse_ten_minute_rss()
        entries = [
            {
                **ten_minute_rss_entries.get(entry["documents_url"], {}),
                **latest_feed_entries.get(entry["documents_url"], {}),
            }
            for entry in {*latest_feed_entries, *ten_minute_rss_entries}
        ]
        for row in entries:
            row["documents_url"] = row["documents_url"].split("edgar/data/")[-1]
            row["cik"] = row["cik"].lstrip("0")
        async with self.engine.begin() as conn:
            existing_document_urls = conn.execute(
                sa.select(daily_rss_table.c.documents_url).where(
                    daily_rss_table.c.documents_url.in_(
                        [row["documents_url"] for row in entries]
                    )
                )
            ).scalars()
        new_entries = [
            row for row in entries if row["documents_url"] not in existing_document_urls
        ]
        await self._daily_rss_upserter.upsert(new_entries)
        # go to documents_url to extract more data.
        self.logger.info(f"Parsing filing pages for {new_entries} new RSS entries.")
        for row in tqdm(new_entries):
            await self._meta_searcher.parse_filing_page(row)

    async def parse_latest_feed(self) -> List[Dict[str, Any]]:
        """Most recent daily. (RSS of https://www.sec.gov/cgi-bin/browse-edgar?action=getcurrent)

        Returns:
            Dict[str, Any]: Filing metadata extracted from the RSS feed.
        """
        url = "https://www.sec.gov/cgi-bin/browse-edgar?action=getcurrent&CIK=&type=&company=&dateb=&owner=include&start=0&count=9999&output=atom"
        feed = await self._get_feed(url)
        entries = []
        if await self._rss_is_updated(feed, url):
            for entry in feed.entries:
                try:
                    # update existing filing data or create new if data does not yet exist for this documents_url
                    filing_data = {}
                    filing_data["documents_url"] = entry["link"]
                    # extract filing type.
                    title_split = entry["title"].split(" - ")
                    filing_data["type"] = title_split[0]
                    name_cik = " - ".join(title_split[1:])
                    cik_match = re.search(r"\((\d{10})\)", name_cik)
                    filing_data["cik"] = cik_match.group(1)
                    filing_data["name"] = name_cik[: cik_match.start()]
                    # parse data from summary HTML format.
                    filed = re.search(
                        r"<b>Filed:</b>\s*(\d{4}-\d{2}-\d{2})", entry.summary
                    )
                    if filed:
                        filing_data["filing_date"] = filed.group(1)
                    entries.append(filing_data)
                except Exception as e:
                    self.logger.error(f"Error parsing {entry}: {e}")
        return entries

    async def parse_ten_minute_rss(self) -> List[Dict[str, Any]]:
        standard_feed_urls = (
            # This is a list of up to 200 of the latest XBRL filings, updated every 10 minutes.
            "https://www.sec.gov/Archives/edgar/xbrlrss.all.xml",
            # Inline XBRL Filings containing financial statements tagged using the US GAAP or IFRS taxonomies in the Inline XBRL format.
            # This is a list of up to 200 of the latest Inline XBRL Filings containing financial statements tagged using the US GAAP or IFRS taxonomies in the Inline XBRL format, updated every 10 minutes.
            "https://www.sec.gov/Archives/edgar/xbrl-inline.rss.xml",
            # This is a list of up to 200 of the latest filings containing financial statements tagged using the US GAAP or IFRS taxonomies, updated every 10 minutes.
            # Filings containing financial statements tagged using the US GAAP or IFRS taxonomies
            "https://www.sec.gov/Archives/edgar/usgaap.rss.xml",
        )
        entries = []
        for url in standard_feed_urls:
            feed = await self._get_feed(url)
            if await self._rss_is_updated(feed, url):
                entries += await parse_standard_feed(feed)
        return entries

    async def _rss_is_updated(self, feed: FeedParserDict, rss_url: str) -> bool:
        """Check is RSS was updated since the last time this function was called.

        Args:
            feed (FeedParserDict): The RSS feed.
            rss_url (str): The URL of the RSS.

        Returns:
            bool: True if the feed has updated, else False.
        """
        if (rss_update_time := feed.feed.updated_parsed) is None:
            self.logger.error(
                "RSS feed at %s does not have field `updated_parsed`.", rss_url
            )
        rss_update_time = datetime(
            rss_update_time.tm_year,
            rss_update_time.tm_mon,
            rss_update_time.tm_mday,
            rss_update_time.tm_hour,
            rss_update_time.tm_min,
            rss_update_time.tm_sec,
            tzinfo=timezone("America/New_York"),
        )
        # TODO convert to UTC.
        async with self.engine.begin() as conn:
            time_last_updated = conn.execute(
                sa.select(rss_last_update_table.c.updated).where(
                    rss_last_update_table.c.rss_url == rss_url
                )
            ).scalar()
            # check is RSS was updated since the last time this function was called.
            if time_last_updated is None or rss_update_time > time_last_updated:
                # save new last updated time.
                conn.execute(
                    sa.update(rss_last_update_table)
                    .values({"updated": rss_update_time})
                    .where(rss_last_update_table.c.rss_url == rss_url)
                )
                self.logger.info("RSS feed at %s is updated.", rss_url)
                return True
        self.logger.info("RSS feed at %s is not updated.", rss_url)
        return False

    async def _get_feed(self, url: str) -> FeedParserDict:
        await rate_limiter.maybe_sleep(url)
        try:
            async with self._http_client.get(url) as resp:
                html = await resp.text()
            return feedparser.parse(html)
        except Exception as e:
            self.logger.error("Could not parse feed at %s: %s", url, e)
