import re
from functools import partial

from fin_data import asset_info as ai
from fin_data.asset_info import CompanyIDType
from fin_data.data.states import abbr_to_state
from fin_data.utils import create_http_client, user_header
from database_utils.upsert_sources import SourcesUpserter
from host_throttle import RateLimiter
from ready_logger import get_logger
from lxml.html import fromstring
from tqdm import tqdm


async def load_search_by_state():
    logger = get_logger(name="sec state")

    UpserterT = partial(SourcesUpserter.create, source="sec-state", logger=logger)
    name_loader = await UpserterT(sa_obj=ai.Name)
    cik_loader = await UpserterT(sa_obj=ai.CIK)
    state_loader = await UpserterT(sa_obj=ai.State)

    rate_limiter = RateLimiter.from_rate_limit(n_requests=1, seconds=1)

    async with create_http_client() as client:
        for state in tqdm(abbr_to_state.keys()):
            # loop over pages of 40
            start = 0
            while True:
                url = f"https://www.sec.gov/cgi-bin/browse-edgar?action=getcompany&State={state}&owner=include&match=&start={start}&count=40&hidefilings=0"
                start += 40
                await rate_limiter.maybe_sleep()
                async with client.get(url, headers=user_header()) as resp:
                    html = await resp.text()
                root = fromstring(html)
                rows = root.xpath('//table[@summary="Results"]//tr')
                if not len(rows):
                    break
                logger.info(f"Extracted {len(rows)} {state} company rows.")
                row_eles = [row.xpath('./td[@scope="row"]') for row in rows]
                row_eles = [eles for eles in row_eles if len(eles) > 2]
                names = [
                    re.sub(r"\s+", " ", str(eles[1].xpath("./text()")[0])).strip()
                    for eles in row_eles
                ]
                ciks = [
                    str(eles[0].xpath("./a/text()")[0]).strip().lstrip("0")
                    for eles in row_eles
                ]
                states = [str(eles[2].xpath("./a/text()")[0]) for eles in row_eles]
                await name_loader.upsert(
                    [
                        {
                            "name": name,
                            "asset_id": cik,
                            "asset_id_type": CompanyIDType.CIK,
                        }
                        for name, cik in zip(names, ciks)
                    ]
                )
                await cik_loader.upsert(
                    [
                        {
                            "cik": cik,
                            "asset_id": name,
                            "asset_id_type": CompanyIDType.NAME,
                        }
                        for name, cik in zip(names, ciks)
                    ]
                )
                await state_loader.upsert(
                    [
                        {
                            "state": state,
                            "asset_id": cik,
                            "asset_id_type": CompanyIDType.CIK,
                        }
                        for state, cik in zip(states, ciks)
                    ]
                )
