import re
from datetime import datetime

import sqlalchemy as sa
from host_throttle import RateLimiter
from html_text import extract_text
from ready_logger import logger
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.asyncio.engine import AsyncEngine
from pydantic import BaseSettings

from .tables import no_matching_cik_table, no_matching_symbol_table

rate_limiter = RateLimiter.from_rate_limit(n_requests=5, seconds=1)


class EmailSettings(BaseSettings):
    sender_email: str
    sender_email_pass: str
    receiver_email: str

    class Config:
        env_prefix = "alerts_"


async def no_matching_cik(html: str, cik: str, engine: AsyncEngine) -> bool:
    if (
        re.search(r"(?i)no\smatching\s(ticker|cik|symbol)", extract_text(html))
        is not None
    ):
        logger.info(f"No matching cik: {cik}")
        statement = (
            postgresql.insert(no_matching_cik_table)
            .values({"cik": cik})
            .on_conflict_do_nothing(index_elements=["cik"])
        )
        async with engine.begin() as conn:
            conn.execute(statement)
        return True
    return False


async def no_matching_symbol(html: str, symbol: str, engine: AsyncEngine) -> bool:
    if (
        re.search(r"(?i)no\smatching\s(ticker|cik|symbol)", extract_text(html))
        is not None
    ):
        logger.info(f"No matching symbol: {symbol}")
        statement = (
            postgresql.insert(no_matching_symbol_table)
            .values({"symbol": symbol})
            .on_conflict_do_nothing(index_elements=["symbol"])
        )
        async with engine.begin() as conn:
            conn.execute(statement)
        return True
    return False


def extract_first(eles):
    return str(eles[0]).strip().strip(",: ") if len(eles) else ""


def comma_filter(text):
    # replace multiple consecutive commas with single comma.
    return re.sub(r",\s*,", ",", text).strip().strip(", ")


def comma_join(eles):
    # join text with comma and space.
    text = ", ".join([str(e).strip() for e in eles])
    text = comma_filter(text)
    # filter whitespace
    text = re.sub(r"\s+", " ", text)
    return text.strip().strip(",: ")


def space_join(eles):
    text = " ".join([str(e).strip() for e in eles])
    text = re.sub(r"\s+", " ", text)
    return text.strip().strip(",: ")
