import asyncio
from logging import Logger
from typing import Any, Dict, Optional

from aiohttp.client import ClientSession
from host_throttle.rate_limiters import RateLimiter, TargetRateLimiters
from host_throttle.rates import ConstantRate
from ready_logger import get_logger
from td_connect import TDAuth


class TDFetch:
    def __init__(self, logger: Logger = None) -> None:
        """Download stuff from TD Ameritrade APIs.

        Args:
            request_sleep (float, optional): Number of seconds to sleep after each requests is sent. Defaults to 1.
        """
        self.logger = logger or get_logger("TDFetch")

        rate = ConstantRate(n_requests=1, seconds=60 * 5)
        self.rate_limiter = TargetRateLimiters(default_limit=rate)

        self._set_clients()

    async def fetch_symbol(
        self, endpoint: str, params: Dict[str, Any] = {}, skip_sleep: bool = False
    ) -> str:
        """Fetch JSON data from an API endpoint.

        Args:
            endpoint (str): The data endpint.
            params (Dict[str, Any], optional): Parameters for GET request. Defaults to {}.

        Returns:
            Any: Response text.
        """
        symbol = params.get("symbol")

        log_meta = symbol or endpoint
        params["apikey"] = self.auth.client_id
        if not skip_sleep:
            pass
            #await self.rate_limiter.maybe_sleep(endpoint, symbol)
        async with self.session.get(
            endpoint,
            headers=self.auth.auth_header,
            params=params,
        ) as response:
            if response.status == 200:
                self.logger.info(f"[{response.status}] {log_meta}")
                return await response.text()
            self.logger.error(f"[{response.status}] {log_meta}")
            if response.status in (401, 403):
                # 401 = Unauthorized, 403 = Forbidden
                await self.reset_api_connections(delay_new=5.5)
                return await self.fetch_symbol(endpoint, params)

    async def reset_api_connections(self, delay_new: Optional[int] = None):
        self.logger.info("Resetting API connections")
        try:
            await self.session.close()
        except Exception as e:
            self.logger.error("Failed to close existing aiohttp session.")
        if delay_new:
            await asyncio.sleep(delay_new)
        self._set_clients()

    def _set_clients(self):
        self.session = ClientSession()
        self.auth = TDAuth()
