import asyncio
import json
import re
import signal
from datetime import datetime
from hashlib import md5
from itertools import count
from typing import Any, Dict, List

from database_utils.column_names import to_safe_snake_case
from database_utils.upsert import Upserter
from fin_data.sources.td_ameritrade.fetch import TDFetch
from fin_data.sources.td_ameritrade.options.tables import (
    td_option_contracts,
    td_option_payloads,
    td_option_quotes,
    td_option_underlyings,
)
from fin_data.sources.td_ameritrade.utils import BlackLister
from ready_logger import get_logger
from tqdm import tqdm


def datetime_from_timestamp_ms(epoch_ms: int):
    return datetime.utcfromtimestamp(epoch_ms / 1000)


def column_name_converter(name: str):
    return re.sub(r"(?i)_in_(long|double)$", "", to_safe_snake_case(name))


class TDOptionChains:
    logger = get_logger(name="TDOptionChains")
    endpoint = "https://api.tdameritrade.com/v1/marketdata/chains"

    @classmethod
    async def create(cls):
        self = cls()
        self.fetcher = TDFetch(self.logger)
        self.blacklister = await BlackLister.create()
        self._underlying_quotes_upserter = await Upserter.create(
            sa_obj=td_option_underlyings,
            on_dupe_update=False,
            values_no_load=["NaN", "", " "],
            column_value_converters={
                "quote_time": datetime_from_timestamp_ms,
                "trade_time": datetime_from_timestamp_ms,
            },
            name_converter=column_name_converter,
            rows_batch_size=700,
            logger=self.logger,
        )
        self._quotes_upserter = await Upserter.create(
            sa_obj=td_option_quotes,
            on_dupe_update=False,
            values_no_load=["NaN", "", " "],
            column_value_converters={
                "quote_time": datetime_from_timestamp_ms,
                "trade_time": datetime_from_timestamp_ms,
            },
            name_converter=column_name_converter,
            rows_batch_size=700,
            logger=self.logger,
        )
        self._contracts_upserter = await Upserter.create(
            sa_obj=td_option_contracts,
            values_no_load=["NaN", "", " "],
            column_value_converters={
                "expiration_date": datetime_from_timestamp_ms,
            },
            name_converter=column_name_converter,
            rows_batch_size=700,
            on_dupe_update=False,
            logger=self.logger,
        )
        self._payload_upserter = await Upserter.create(
            sa_obj=td_option_payloads,
            rows_batch_size=200,
            on_dupe_update=False,
            logger=self.logger,
        )
        self._db_load_tasks = []
        loop = asyncio.get_running_loop()
        signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        for s in signals:
            loop.add_signal_handler(
                s, lambda s=s: asyncio.create_task(self.shutdown(s))
            )
        return self

    async def fetch_quote(
        self, symbol: str, skip_sleep: bool = False
    ) -> Dict[str, Any]:
        await asyncio.sleep(0.5)
        return await self.fetcher.fetch_symbol(
            endpoint=self.endpoint,
            params={
                "strikeCount": 100,
                "includeQuotes": "true",
                "symbol": symbol,
            },
            skip_sleep=skip_sleep,
        )

    def parse_response(self, response_text: str):
        response = json.loads(response_text)
        if (status := response.pop("status", None)) not in ("SUCCESS", None):
            self.logger.error(f"Response has non-SUCCESS status: {status}")
            return False

        missing = [k for k in ("callExpDateMap", "putExpDateMap") if k not in response]
        if len(missing):
            raise RuntimeError(
                f"Response is missing expected key(s): {missing}. Is this running outside of trading hours?"
            )

        chains = [
            contract
            for contract_type in ("callExpDateMap", "putExpDateMap")
            for date_data in response[contract_type].values()
            for strike_price_data in date_data.values()
            for contract in strike_price_data
            if contract.get("quoteTimeInLong")
        ]

        # extract option underlying data.
        underlying = response["underlying"]
        underlying["id"] = underlying_id = md5(
            "".join([underlying["symbol"], str(underlying["quoteTime"])]).encode()
        ).hexdigest()

        for c in chains:
            c["underlying_symbol"] = response["symbol"]
            c["underlying_quote_id"] = underlying_id
        
        # add to buffers.
        self._payload_upserter.buffer.append(
            {
                "payload": response,
                "payload_id": md5(response_text.encode()).hexdigest(),
            }
        )
        self._underlying_quotes_upserter.buffer.append(underlying)
        self._quotes_upserter.buffer.extend(chains)
        self._contracts_upserter.buffer.extend(chains)
        return True

    async def stream(self, symbols: List[str], db_batch_size: int = 7):
        n_processed = count()
        while True:
            self.logger.info(f"Processing {len(symbols)} symbols")
            for symbol in tqdm(symbols):
                # fetch symbol, allowing 1 retry.
                for t in range(2):
                    try:
                        response = await self.fetch_quote(symbol, skip_sleep=t > 0)
                        if self.parse_response(response):
                            self.blacklister.mark_ok(self.endpoint, symbol)
                            break
                    except Exception as e:
                        self.logger.error(f"Error processing {symbol}: {e}")
                    if await self.blacklister.mark_bad(self.endpoint, symbol):
                        self.logger.warning(f"Removing symbol: {symbol}")
                        symbols.remove(symbol)
                        break
                if next(n_processed) % db_batch_size == 0:
                    self.logger.info("Creating database task batch.")
                    # load response batch.
                    self._db_load_tasks.append(
                        asyncio.create_task(self._underlying_quotes_upserter.upsert())
                    )
                    self._db_load_tasks.append(asyncio.create_task(self._quotes_upserter.upsert()))
                    self._db_load_tasks.append(
                        asyncio.create_task(self._contracts_upserter.upsert())
                    )
                    self._db_load_tasks.append(asyncio.create_task(self._payload_upserter.upsert()))
                    self._db_load_tasks = [t for t in self._db_load_tasks if not t.done()]
                    self.logger.info(
                        f"{len(self._db_load_tasks)} outstanding database load tasks."
                    )


    async def shutdown(self, signal=None):
        print(
            f"Received shutdown signal {signal}. Waiting on {len(self._db_load_tasks)} database tasks."
        )
        await asyncio.gather(*self._db_load_tasks)
        


async def main():
    from pathlib import Path

    from database_utils.utils import get_engine
    from dotenv import load_dotenv

    load_dotenv("/home/dan/.env")

    e = get_engine()
    #s = await theotrade_options_symbols(e)
    with Path.home().joinpath('top_options.csv').open(mode='r') as f:
        s = [l.strip('\n') for l in f]

    l = await TDOptionChains.create()
    await l.stream(s, 7)


if __name__ == "__main__":
    asyncio.run(main())
