from datetime import datetime
from enum import Enum

import sqlalchemy as sa
from fin_data.sources.td_ameritrade.tables import metadata
from sqlalchemy.dialects.postgresql import ENUM, JSONB

td_option_payloads = sa.Table(
    "option_payloads",
    metadata,
    sa.Column("payload_id", sa.String, primary_key=True),
    sa.Column("payload", JSONB, nullable=False),
    sa.Column(
        "inserted", sa.DateTime(timezone=False), default=datetime.utcnow, nullable=False
    ),
)

td_option_underlyings = sa.Table(
    "option_underlying_quotes",
    metadata,
    sa.Column(
        "id",
        sa.String,
        primary_key=True,
        comment="ID generated from hash of symbol + quote_time",
    ),
    sa.Column("symbol", sa.String, nullable=False),
    sa.Column("quote_time", sa.DateTime(timezone=False), nullable=False),
    sa.Column("trade_time", sa.DateTime(timezone=False)),
    sa.Column("bid", sa.Float),
    sa.Column("ask", sa.Float),
    sa.Column("close", sa.Float),
    sa.Column("change", sa.Float),
    sa.Column("percent_change", sa.Float),
    sa.Column("last", sa.Float),
    sa.Column("mark", sa.Float),
    sa.Column("mark_change", sa.Float),
    sa.Column("mark_percent_change", sa.Float),
    sa.Column("bid_size", sa.Integer),
    sa.Column("ask_size", sa.Integer),
    sa.Column("high_price", sa.Float),
    sa.Column("low_price", sa.Float),
    sa.Column("open_price", sa.Float),
    sa.Column("total_volume", sa.Integer),
    sa.Column("fifty_two_week_high", sa.Float),
    sa.Column("fifty_two_week_low", sa.Float),
    sa.Column("delayed", sa.Boolean),
)

class PutCall(Enum):
    PUT = "PUT"
    CALL = "CALL"

td_option_contracts = sa.Table(
    "option_contracts",
    metadata,
    sa.Column("symbol", sa.String, primary_key=True),
    sa.Column(
        "underlying_symbol",
        sa.String,
        index=True,
        nullable=False,
        comment="Need the underlying symbol to link to underlying quotes.",
    ),
    sa.Column("mini", sa.Boolean),
    sa.Column("is_index_option", sa.Boolean),
    sa.Column("put_call", ENUM(PutCall)),
    sa.Column("strike_price", sa.Float),
    sa.Column("expiration_date", sa.DateTime(timezone=False)),
    sa.Column("days_to_expiration", sa.Integer),
    # 'Q' = quarterly, 'S' = weekly, 'R' = monthly
    sa.Column("expiration_type", sa.String),
    sa.Column("exchange", sa.String),
    sa.Column("exchange_name", sa.String),
    sa.Column("settlement_type", sa.String),
    sa.Column("penny_pilot", sa.Boolean),
    sa.Column("non_standard", sa.Boolean),
)

td_option_quotes = sa.Table(
    "option_quotes",
    metadata,
    sa.Column("symbol", sa.String, primary_key=True),
    sa.Column(
        "underlying_quote_id",
        sa.String,
        index=True,
        nullable=False,
        comment="Link to row in option_underlying table.",
    ),
    sa.Column(
        "quote_time",
        sa.DateTime(timezone=False),
        primary_key=True,
        # comment="HypertablePartiton",
    ),
    sa.Column("trade_time", sa.DateTime(timezone=False)),
    sa.Column("bid", sa.Float),
    sa.Column("bid_size", sa.Integer),
    sa.Column("ask", sa.Float),
    sa.Column("ask_size", sa.Integer),
    sa.Column("last", sa.Float),
    sa.Column("last_size", sa.Integer),
    sa.Column("percent_change", sa.Float),
    sa.Column("mark", sa.Float),
    sa.Column("mark_change", sa.Float),
    sa.Column("mark_percent_change", sa.Float),
    sa.Column("high_price", sa.Float),
    sa.Column("low_price", sa.Float),
    sa.Column("open_price", sa.Float),
    sa.Column("close_price", sa.Float),
    sa.Column("total_volume", sa.Integer),
    sa.Column("net_change", sa.Float),
    sa.Column("volatility", sa.Float),
    # TODO check this is constant for all symbols.
    sa.Column("theoretical_volatility", sa.Float),
    sa.Column("intrinsic_value", sa.Float),
    sa.Column("delta", sa.Float),
    sa.Column("gamma", sa.Float),
    sa.Column("theta", sa.Float),
    sa.Column("vega", sa.Float),
    sa.Column("rho", sa.Float),
    sa.Column("open_interest", sa.Integer),
    sa.Column("time_value", sa.Float),
    sa.Column("theoretical_option_value", sa.Float),
    sa.Column("in_the_money", sa.Boolean),
)

"""
https://fc.wealthmsi.com/sites/client/tdanl/TDANewsletter/article.vm?topic=6483&siteContent=9291
On January 26, 2007, the options exchanges kicked off a penny pricing pilot program that included 13 stocks and Exchange Traded Funds (ETFs) with options listed at one-cent increments. 
Since that time, the list has been expanded to more than 100. 
These one-cent increments are another important trend that has put more power into the hands of individual investors. 
No longer are we fighting 5-, 10- or 25-cent spreads between bid and ask prices; 
the spread on many of the most actively traded names is often just one or two pennies.
"""
