import csv
import re
from datetime import datetime
from pathlib import Path
from typing import List, Tuple

import pandas as pd


def format_column(name: str) -> str:
    return re.sub(r"\s+", "_", name).lower().replace(".", "_").strip()


def extract_symbol_and_time(rows: List[List[str]]) -> Tuple[str, datetime]:
    for row in rows:
        if len(row) and (
            match := re.match(
                r"Stock quote and option quote for ([^\s]+) on (\d{1,2}\/\d{1,2}\/\d{2} \d{2}:\d{2}:\d{2})$",
                row[0].strip(),
            )
        ):
            symbol, quote_time = match.group(1), match.group(2)
            return symbol, datetime.strptime(quote_time, "%m/%d/%y %H:%M:%S")


def extract_underlying(rows: List[List[str]]) -> pd.DataFrame:
    """Parse the 'underlying' info from the top of the exported file.
    Expected format (with random example values):
    UNDERLYING
    LAST,LX,Net Chng,BID,BX,ASK,AX,Size,Volume,Open,High,Low
    418.10,Z,+.70,418.10,Q,418.11,Q,18 x 2,"25,754,415",420.32,420.72,417.54
    """
    for idx, row in enumerate(rows):
        if len(row) and row[0].strip() == "UNDERLYING":
            return pd.DataFrame([rows[idx + 2]], columns=[f'u_{format_column(c)}' for c in rows[idx + 1]])


def extract_underlying_extra(rows: List[List[str]]) -> pd.DataFrame:
    """Parse the 'underlying extra info' info from the top of the exported file.
    Expected format (with random example values):
    UNDERLYING EXTRA INFO
    Last Size,Yield,PE,Div,Div.Freq,Ex Div.Date,52High,52Low,Shares,Beta
    100,1.22%,<empty>,1.278,Q,3/19/21,420.72,272.99,"867,632,116",.9952
    """
    for idx, row in enumerate(rows):
        if len(row) and row[0].strip() == "UNDERLYING EXTRA INFO":
            return pd.DataFrame([rows[idx + 2]], columns=[f'u_{format_column(c)}' for c in rows[idx + 1]])


def extract_contracts(rows: List[List[str]]) -> pd.DataFrame:
    """Extract data from the main file body."""
    rows = iter(rows)
    row_dicts = []
    # loop through and process all rows in file.
    while True:
        try:
            row = next(rows)
            if len(row):
                # check is line is start of a new contract.
                match = re.search(
                    r"(\d{1,2} [A-Z]{3,} \d{1,2})\s+\((\d+)\)\s+(\d+)\s*(\(([a-zA-Z]+)\))?",
                    row[0],
                )
                if match is not None:
                    # row should be one price of text, no commas.
                    assert len(row) == 1
                    exp = match.group(1)
                    contract_data = {
                        "days_till_expiration": match.group(2),
                        "lot_size": match.group(3),
                        "duration": match.group(5),
                    }
                    # section header is next line.
                    header = next(rows)
                    # call and put columns should be separated by 'Exp', 'Strike'
                    exp_idx = header.index("Exp")
                    assert header[exp_idx + 1] == "Strike"
                    # rename columns.
                    call_columns = [f"c_{format_column(c)}" for c in header[:exp_idx]]
                    put_columns = [f"p_{format_column(c)}" for c in header[exp_idx + 2 :]]
                    header = call_columns + ["expiration_date", "strike"] + put_columns
                    # continue until blank line.
                    while len(row):
                        # convert row list to dict.
                        if row := dict(zip(header, next(rows))):
                            assert row["expiration_date"] == exp
                            row_dicts.append({**row, **contract_data})
        except StopIteration:
            # reached end of file. return data.
            df = pd.DataFrame(row_dicts)
            # set column order.
            ordered_columns = [
                "expiration_date",
                "days_till_expiration",
                "duration",
                "strike",
            ]
            df = pd.concat(
                [
                    df[ordered_columns],
                    df[[c for c in df.columns if c not in ordered_columns]],
                ],
                axis=1,
            )
            return df


def clean_df(df):
    drop_columns = [c for c in [
        'c_size', 'p_size',
        'c_bx', 'p_bx',
        'c_ax', 'p_ax',
        'c_lx', 'p_lx',
        'c_option_code', 'p_option_code',
    ] if c in df]
    df.drop(drop_columns, axis=1, inplace=True)
    df.replace(["", "<empty>", "--"], None, inplace=True)
    # drop rows that are completely empty.
    df.dropna(axis=0, how="all", inplace=True)
    # drop columns that are completely empty.
    df.dropna(axis=1, how="all", inplace=True)
    # convert date strings to datetime objects.
    for c in ('expiration_date','u_ex_div_date'):
        if c in df:
            df[c] = pd.to_datetime(df[c])
    # remove percent signs for values.
    for c in df.columns:
        if df[c].dtype == object and df[c].str.endswith("%").all():
            df[c] = df[c].str.rstrip("%")
            if not "%" in c:
                df.rename(columns={c: f"{c}(%)"}, inplace=True)
    # format remove commas and leading '+' from numbers.    
    for c in df.columns:
        if df[c].dtype == object:
            df[c] = df[c].apply(
                lambda x: x.replace(",", "").lstrip("+")
                if x and re.match(r"^\+?\d[\d,.]+$", x)
                else x
            ) 


def save_df(
    df: pd.DataFrame,
    save_dir: Path,
    symbol: str,
    quote_time: str,
    data_name: str,
    save_format: str,
):
    save_path = save_dir / f"{symbol}_{data_name}_{quote_time}.{save_format}"
    # save to file.
    getattr(df, f"to_{save_format}")(save_path, index=False)


def parse_export_file(file_path, save_format="csv", save_dir=None):
    """Pase file sections and save to CSVs."""
    # load rows from export file.
    with file_path.open(mode="r") as f:
        rows = list(csv.reader(f))

    symbol, quote_time = extract_symbol_and_time(rows)
    quote_time = quote_time.strftime("%a_%b%d_%y__%H_%M_%S")

    save_dir = (save_dir or file_path.parent) / f"{symbol}_{quote_time}"
    save_dir.mkdir(parents=True, exist_ok=True)

    # parse underlying.
    underlying_df = pd.concat([extract_underlying(rows), extract_underlying_extra(rows)], axis=1)
    clean_df(underlying_df)
    save_df(underlying_df, save_dir, symbol, quote_time, "underlying", save_format)

    # parse contracts.
    contracts_df = extract_contracts(rows)
    clean_df(contracts_df)
    save_df(contracts_df, save_dir, symbol, quote_time, "contracts", save_format)


if __name__ == "__main__":
    import sys

    file = (
        Path.home() / "2022-04-02-StockAndOptionQuoteForTSLA.csv" #"2022-04-03-StockAndOptionQuoteForSPY.csv"
    )

    parse_export_file(file)
