import json
from functools import partial
from typing import Any
from urllib.parse import urlparse

from fin_data import asset_info as ai
from fin_data.sources import random_user
from fin_data.utils import create_http_client, extract_column_name
from database_utils.upsert_sources import SourcesUpserter
from ready_logger import get_logger


class YahooFundamentals:
    logger = get_logger("YahooFinance")
    UpserterT = partial(SourcesUpserter.create, source="yahoo-finance", logger=logger)

    @classmethod
    async def create(cls):
        self = cls()
        self.http_client = create_http_client()
        self.column_value_upserters = {
            field: {
                "upserter": await self.UpserterT.create(entity),
                "column_name": extract_column_name(entity),
            }
            for field, entity in {
                "business_address": ai.BusinessAddress,
                "city": ai.City,
                "country": ai.Country,
                "fax": ai.Fax,
                "companyOfficers": ai.CompanyOfficers,
                "fullTimeEmployees": ai.FullTimeEmployees,
                "industry": ai.Industry,
                "longBusinessSummary": ai.Summary,
                "phone": ai.Phone,
                "sector": ai.Sector,
                "state": ai.State,
                "website": ai.Website,
                "zip": ai.ZipCode,
                "longName": ai.Name,
                "exchange": ai.Exchange,
                "exchangeName": ai.Exchange,
                "logo": ai.Logo,
                "currencySymbol": ai.CurrencySymbol,
                "currency": ai.Currency,
            }.items()
        }
        self.has_stock_upserter = await self.UpserterT.create(ai.HasStock)
        self.has_etf_upserter = await self.UpserterT.create(ai.HasETF)

    def fetch_fundamentals(self, symbol: str):
        """Keys: ['defaultKeyStatistics', 'details', 'summaryProfile', 'recommendationTrend', 'financialsTemplate',
        'earnings', 'price', 'financialData', 'quoteType', 'calendarEvents', 'summaryDetail', 'symbol',
        'esgScores', 'upgradeDowngradeHistory', 'pageViews']
        """
        data = self._parse_response(f"https://finance.yahoo.com/quote/{symbol}")
        if data:
            website = data.get("summaryProfile", {}).get("website")
            if website:
                domain = urlparse(website).netloc.split("www.")[-1]
                data["summaryProfile"]["logo"] = f"https://logo.clearbit.com/{domain}"
            return data

    async def load_fundamentals_data(self, symbol: str):
        data = self.fetch_fundamentals(symbol)
        if data:
            data = {**data.get("price", {}), **data.get("summaryProfile", {})}
            if "symbol" not in data:
                data["symbol"] = symbol
        asset_id = {
            "asset_id": data["symbol"],
            "asset_id_type": ai.CompanyIDType.SYMBOL,
        }
        if "companyOfficers" in data:
            data["companyOfficers"] = json.dumps(data["companyOfficers"])
        if quote_type := data.get("quoteType"):
            if quote_type == "ETF":
                await self.has_etf_upserter.upsert(asset_id)
            elif quote_type == "commonStock":
                await self.has_stock_upserter.upsert(asset_id)
            else:
                raise RuntimeError("Unknown quote type: %s", quote_type)
        address = []
        if "address1" in data:
            address.append(data["address1"])
        if "address2" in data:
            address.append(data["address2"])
        if len(address):
            data["business_address"] = json.dumps(address)
        for field, db in self.column_value_upserters.items():
            if field in data:
                await db["upserter"].upsert(
                    {**asset_id, db["column_name"]: data[field]}
                )

    def holders(self, symbol: str):
        """Keys: ['financialsTemplate', 'majorDirectHolders', 'price', 'fundOwnership', 'insiderTransactions',
        'insiderHolders', 'netSharePurchaseActivity', 'majorHoldersBreakdown', 'quoteType',
        'institutionOwnership', 'summaryDetail', 'symbol', 'pageViews']
        """
        return self._parse_response(f"https://finance.yahoo.com/quote/{symbol}/holders")

    async def _parse_response(self, url: str) -> Any:
        try:
            async with self.http_client.get(
                url, headers={"User-Agent": random_user()}
            ) as resp:
                html = await resp.text()

                json_str = (
                    html.split("root.App.main =")[1]
                    .split("(this)")[0]
                    .split(";\n}")[0]
                    .strip()
                )
                return json.loads(json_str)["context"]["dispatcher"]["stores"][
                    "QuoteSummaryStore"
                ]
        except Exception as e:
            self.logger.error(f"Error Loading data for {url}: {e}")
