"""Add manually selected symbols or symbols from export files."""

import argparse
import asyncio
import csv

from database_utils.upsert_sources import SourcesUpserter
from fin_data.asset_info import symbols_table
from ready_logger import logger


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--source-name', '-n', type=str, required=True, help='Name of the source that the symbols are from.')
    parser.add_argument('--symbols', '-s', type=str, nargs='+', help='Symbols that should be upserted.')
    parser.add_argument('--file', '-f', type=str, help='File containing symbols that should be upserted.')
    parser.add_argument('--file-fmt', '-t', choices=['tos', 'txt'], default='txt', help='Format of the symbols file. (tos=Thinkorswim export. txt=line delimited text file.)')
    parser.add_argument('--bools', '-b', nargs='+', help='Boolean columns that should be set to true.')
    return parser.parse_args()

async def upsert_symbols():
    args = parse_args()
    if not args.file and not args.symbols:
        raise ValueError("Either `symbols` or `file` must be provided.")
    symbols = set()
    if args.file:
        logger.info(f"Reading symbols from file {args.file}")
        if args.file_fmt == 'tos':
            with open(args.file, mode='r') as f:
                symbols.update([r[0].split('2')[0].split('.')[1] for r in csv.reader(f)])
        elif args.file_fmt == 'txt':
            with open(args.file, 'r') as f:
                symbols.update([r.strip() for r in f])
        else:
            raise ValueError(f'Unknown file format: {args.file_fmt}')
    if args.symbols:
        logger.info(f"Adding {len(args.symbols)} manually entered symbols.")
        symbols.update(args.symbols)

    upserter = await SourcesUpserter.create(sa_obj=symbols_table, source=args.source_name)
    rows = [{'symbol': symbol} for symbol in symbols]
    if args.bools:
        bools = {column: True for column in args.bools}
        for row in rows:
            row.update(bools)
    logger.info(f"Upseriting {len(rows)} symbols with {len(bools)} True bool columns: {bools}")
    await upserter.upsert(rows)
    

# Poetry script installer needs non-async function.
def run():
    asyncio.run(upsert_symbols())


if __name__=='__main__':
    run()
